import json
import os

import mock

from mercury.common.helpers.cli import CLIResult
from mercury_agent.hardware.drivers.megaraid import MegaRAIDActions, \
    MegaRaidSASDriver
from mercury_agent.hardware.raid.abstraction.api import RAIDAbstractionException

from ..base import MercuryAgentUnitTest


def get_storcli_dall_show(c):
    """ Gets data from json resource

    :param c: unused
    :return:
    """
    del c
    with open(os.path.join(os.path.dirname(__file__),
                           '../resources/storcli_dall_show.json')) as fp:
        data = json.load(fp)

        controller_list = []
        for controller in data['Controllers']:
            # Account for some crazy JSON schema
            controller_list.append(
                controller['Response Data']['Response Data']
            )

        return controller_list


def get_controllers():
    """ Gets controller information from json resource
    :return:
    """
    with open(os.path.join(os.path.dirname(__file__),
                           '../resources/storcli.json')) as fp:
        return [c['Response Data'] for c in json.load(fp)['Controllers']]


class DummyMegaRAIDActions(MegaRAIDActions):
    """ A dummy megaraid actions driver for easier patching """

    def __init__(self):
        super(MegaRAIDActions, self).__init__()
        self.storcli = mock.Mock()

        # controllers is a property
        self.storcli.controllers = get_controllers()

        self.storcli.get_disk_group = get_storcli_dall_show


class TestMegaRAIDActions(MercuryAgentUnitTest):
    """ MegaRAIDActions Test Case """

    def setUp(self):
        """ Instantiates a dummy module for use in test methods """
        super(TestMegaRAIDActions, self).setUp()

        self.dummy_actions = DummyMegaRAIDActions()

    def test_transform_configuration(self):
        """ Test overall transform operations """
        configuration = self.dummy_actions.transform_configuration(0)
        self.assertTrue(len(configuration['arrays']) == 2)
        self.assertEqual(configuration['arrays'][0]['free_space'], 0)
        self.assertEqual(
            configuration['arrays'][1]['physical_drives'][1]['extra']
            ['address'], '32:2')

        self.assertEqual(
            MegaRAIDActions.get_array_index_by_dg(
                configuration['arrays'], 1), 1)

        self.assertEqual(
            MegaRAIDActions.get_array_index_by_dg(
                configuration['arrays'], 99), -1)

    def test_without_available_drives(self):
        dg_info = get_storcli_dall_show(0)

        del dg_info[0]['UN-CONFIGURED DRIVE LIST']
        new_dummy_actions = DummyMegaRAIDActions()
        new_dummy_actions.storcli.get_disk_group = mock.Mock()
        new_dummy_actions.storcli.get_disk_group.return_value = dg_info

        configuration = new_dummy_actions.transform_configuration(0)
        self.assertFalse(configuration['unassigned'])
        self.assertFalse(configuration['spares'])

    def test_with_global_hotspare(self):
        """ Convert stored data so that it contains a GHS"""
        dg_info = get_storcli_dall_show(0)
        for drive in dg_info[0]['UN-CONFIGURED DRIVE LIST']:
            if drive['State'] == 'DHS':
                drive['State'] = 'GHS'
                drive['DG'] = '-'

        new_dummy_actions = DummyMegaRAIDActions()
        new_dummy_actions.storcli.get_disk_group = mock.Mock()
        new_dummy_actions.storcli.get_disk_group.return_value = dg_info

        configuration = new_dummy_actions.transform_configuration(0)

        self.assertEqual(configuration['spares'][0]['extra']['spare_type'],
                         'global')

    def test_with_no_hotspare(self):
        dg_info = get_storcli_dall_show(0)
        for drive in dg_info[0]['UN-CONFIGURED DRIVE LIST']:
            if drive['State'] in DummyMegaRAIDActions.hotspare_map:
                drive['State'] = 'UGood'  # UGood, bro?
                drive['DG'] = '-'
        new_dummy_actions = DummyMegaRAIDActions()
        new_dummy_actions.storcli.get_disk_group = mock.Mock()
        new_dummy_actions.storcli.get_disk_group.return_value = dg_info
        configuration = new_dummy_actions.transform_configuration(0)
        self.assertFalse(configuration['spares'])

    def test_transform_adapter_info(self):
        adapter_info = self.dummy_actions.transform_adapter_info(0)
        self.assertEqual(adapter_info['name'], 'PERC 6/i Integrated')
        self.assertEqual(MegaRAIDActions.get_controller_id(adapter_info),
                         0)
        self.assertRaises(RAIDAbstractionException,
                          self.dummy_actions.transform_adapter_info, *(100,))

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    @mock.patch('mercury_agent.hardware.drivers.megaraid.get_configuration')
    def test_real_init(self, mock_get_configuration, mock_cli):
        """ Tests real class __init__ method

        :param mock_get_configuration:
        :param mock_cli:
        :return:
        """
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        mock_get_configuration.return_value = {}

        assert MegaRAIDActions()

    def test_get_vendor_info_static(self):
        """ Tests get_vendor_info static method"""
        adapter = {
            'Basics': {},
            'Version': {},
            'Bus': {},
            'Status': {},
            'Supported Adapter Operations': {},
            'Supported PD Operations': {},
            'Supported VD Operations': {}
        }
        result = MegaRAIDActions.get_vendor_info(adapter)

        keys = [
            'general',
            'version_info',
            'bus',
            'status',
            'supported_adapter_ops',
            'supported_pd_ops',
            'supported_vd_ops',
            'bbu_info',
        ]

        missing = []
        for key in keys:
            if key not in result:
                missing.append(key)

        self.assertFalse(missing, 'Missing keys in output {}'.format(missing))

    def test_sort_drives_static(self):
        """ Test drive sorting  """
        drives = [
            {
                'extra':
                    {
                        'drive_id': 10
                    }
            },
            {
                'extra':
                    {
                        'drive_id': 5
                    }
            },
            {
                'extra':
                    {
                        'drive_id': 7
                    }
            },
            {
                'extra':
                    {
                        'drive_id': 0
                    }
            },
        ]

        MegaRAIDActions.sort_drives(drives)

        self.assertEqual(drives[0]['extra']['drive_id'], 0)
        self.assertEqual(drives[1]['extra']['drive_id'], 5)
        self.assertEqual(drives[2]['extra']['drive_id'], 7)
        self.assertEqual(drives[3]['extra']['drive_id'], 10)

    def test_create(self):
        """ Test create implementation """
        # Data does not have available drives, create one
        dg_info = get_storcli_dall_show(0)

        for drive in dg_info[0]['UN-CONFIGURED DRIVE LIST']:
            drive['State'] = 'UGood'
            drive['DG'] = '-'

        new_dummy_actions = DummyMegaRAIDActions()
        new_dummy_actions.storcli.get_disk_group = mock.Mock()
        new_dummy_actions.storcli.get_disk_group.return_value = dg_info

        adapter_info = new_dummy_actions.transform_adapter_info(0)
        drives = new_dummy_actions.get_drives_from_selection(0, [3])

        new_dummy_actions.create(adapter_info, 0, drives=drives)

        for drive in drives:
            drive['extra']['vendor_state'] = 'UBad'

        self.assertRaises(RAIDAbstractionException, new_dummy_actions.create,
                          *(adapter_info, 0, drives))

    def test_delete_logical_drive(self):
        """ Test delete_logical_drive implementation """
        self.dummy_actions.storcli.delete = mock.Mock()

        self.dummy_actions.delete_logical_drive(0, 0, 0)

        self.dummy_actions.storcli.delete.assert_called_with(**{
            'controller': 0,
            'virtual_drive': 0
        })

        self.assertRaises(RAIDAbstractionException,
                          self.dummy_actions.delete_logical_drive,
                          *(0, 0, 99))

    def test_clear_configuration(self):
        self.dummy_actions.storcli.delete = mock.Mock()
        self.dummy_actions.clear_configuration(0)
        self.dummy_actions.storcli.delete.assert_called_with(
            **{
                'controller': 0,
                'virtual_drive': 'all'
            }
        )

    def test_add_spares(self):
        """ Test create implementation """
        # Data does not have available drives, create one
        dg_info = get_storcli_dall_show(0)

        for drive in dg_info[0]['UN-CONFIGURED DRIVE LIST']:
            drive['State'] = 'UGood'
            drive['DG'] = '-'

        new_dummy_actions = DummyMegaRAIDActions()
        new_dummy_actions.storcli.get_disk_group = mock.Mock()
        new_dummy_actions.storcli.get_disk_group.return_value = dg_info

        new_dummy_actions.storcli.add_hotspare = mock.Mock()

        new_dummy_actions.add_spares(0, [3])

        new_dummy_actions.storcli.add_hotspare.assert_called_with(**{
            'controller': 0,
            'enclosure': 32,
            'slot': 3,
            'disk_groups': []
        })

        new_dummy_actions.add_spares(0, [3], [0])

        new_dummy_actions.storcli.add_hotspare.assert_called_with(**{
            'controller': 0,
            'enclosure': 32,
            'slot': 3,
            'disk_groups': [0]
        })

        self.assertRaises(RAIDAbstractionException,
                          new_dummy_actions.add_spares,
                          *(0, [3], [99]))


class MercuryMegaRaidDSDriverTest(MercuryAgentUnitTest):
    def setUp(self):
        super(MercuryMegaRaidDSDriverTest, self).setUp()
        self.pci_data = [{'class_id': '0104',
                          'class_name': 'RAID bus controller',
                          'device_id': '005d',
                          'device_name': 'MegaRAID SAS-3 3108 [Invader]',
                          'driver': 'megaraid_sas',
                          'revision': '02',
                          'sdevice_id': '1f49',
                          'sdevice_name': 'PERC H730 Mini',
                          'slot': '02:00.0',
                          'svendor_id': '1028',
                          'svendor_name': 'Dell',
                          'vendor_id': '1000',
                          'vendor_name': 'LSI Logic / Symbios Logic'}]

    def test_probe(self):
        self.assertEqual(MegaRaidSASDriver.probe(self.pci_data), ['02:00.0'])

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    @mock.patch('mercury_agent.hardware.drivers.megaraid.get_configuration')
    def test_inspect(self, mock_cli, mock_get_configuration):
        driver = MegaRaidSASDriver(['02:00.0'])
        driver.handler = mock.Mock()
        driver.handler.get_adapter_info = mock.Mock()

        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        mock_get_configuration.return_value = {}

        driver.inspect()

        driver.handler.get_adapter_info.assert_called_with(0)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury_agent.capabilities import capability
from mercury_agent.configuration import get_configuration
from mercury_agent.inspector import inspect
from mercury_agent.inspector.inspect import global_device_info
from mercury_agent.inspector.inspectors import health


@capability('inspector', description='Run inspector')
def inspector():
    """
    Manually run inspectors
    :return: results
    """
    return inspect.inspect()


@capability('check_hardware', description='Check hardware for errors')
def check_hardware():
    """
    Checks hardware for inconsistencies and defects. Returns a list of discovered critical errors.
    :return:
    """
    configuration = get_configuration().agent
    errors = []
    _health_data = health.system_health_inspector(global_device_info)
    if _health_data['corrected_hardware_event_count'] >= configuration.hardware.mce_threshold:
        errors.append(
            'MCE count is {} which is above the configured threshold of {}'.format(
                _health_data['corrected_hardware_event_count'],
                configuration.hardware.mce_threshold))

    return {
        'errors': errors,
        'error_count': len(errors)
    }



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import sys
import traceback

log = logging.getLogger(__name__)

inspectors = []
late_inspectors = []
async_inspectors = []

# TODO: Consider manifest driven inspector orchestration


def run_inspector(name, f, *args, **kwargs):
    # noinspection PyBroadException
    try:
        log.debug('Running inspector: %s' % name)
        return f(*args, **kwargs)
    except Exception:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        path, line, scope, code = traceback.extract_tb(exc_traceback)[-1]
        log.error('Inspector raised an unhandled exception (%s : %s): name=%s, scope=%s, '
                  'path=%s, line=%d, code=%s' % (exc_type,
                                                 exc_value,
                                                 name,
                                                 scope,
                                                 path,
                                                 line,
                                                 code))
        return None


def expose(name):
    def wrap(f):
        def wrapped_f(*args, **kwargs):
            return run_inspector(name, f, *args, **kwargs)
        log.debug('Adding runtime inspector %s (%s)' % (f.__name__, name))
        wrapped_f.__name__ = f.__name__
        wrapped_f.__doc__ = f.__doc__
        inspectors.append((name, wrapped_f))
        return wrapped_f
    return wrap


def expose_late(name, run_if=None):
    """Hardware dependent inspectors, such as those dependent on OEM/ODM utilities
    :param run_if: callback funtion that takes device_info as an argument. This is optional,
        run_if can always be added by other means
    """
    def wrap(f):
        def wrapped_f(early_device_info):
            if hasattr(f, 'run_if') and not f.run_if(early_device_info):
                log.info('Requirement not satisfied for %s (%s)' % (f.__name__, name))
                return None
            return run_inspector(name, f, early_device_info)
        log.debug('Adding late inspector %s (%s)' % (f.__name__, name))
        wrapped_f.__name__ = f.__name__
        wrapped_f.__doc__ = f.__doc__
        if run_if:
            wrapped_f.run_if = run_if
        late_inspectors.append((name, wrapped_f))
        return wrapped_f
    return wrap



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging

from mercury_agent.configuration import get_configuration
from mercury_agent.hardware.drivers import DriverBase, driver
from mercury_agent.hardware.obm.ipmitool import (
    IPMIToolDrac,
    IPMIToolHP,
)

from mercury_agent.hardware.obm.racadm import SimpleRAC


log = logging.getLogger(__name__)


class IPMIDriverBase(DriverBase):
    name = 'ipmi'
    driver_type = 'bmc'
    wants = ''

    @classmethod
    def probe(cls, context_data):
        raise NotImplementedError

    def inspect(self):
        return {
            'network': self.handler.net_info,
            'system': self.handler.bmc_info
        }


@driver()
class HPILODriver(IPMIDriverBase):
    name = 'ilo'
    _handler = IPMIToolHP
    wants = 'pci'

    PCI_DEVICE_IDS = [
        "3306"  # Integrated Lights-Out Standard Slave Instrumentation & System
                # Support
    ]

    @classmethod
    def probe(cls, context_data):
        for pci_device in context_data:
            for my_device_id in cls.PCI_DEVICE_IDS:
                if pci_device['device_id'] == my_device_id:
                    return True
        return False


@driver()
class DRACDriver(IPMIDriverBase):
    name = 'drac'
    _handler = IPMIToolDrac
    wants = 'dmi'

    @classmethod
    def probe(cls, context_data):
        """ Tests that the server is a dell variant and has a compatible DRAC
        interface

        :param context_data: DMI information
        :return:
        """
        if context_data['sys_vendor'] == 'Dell Inc.':
            simple_rac = SimpleRAC(
                get_configuration().agent.hardware.obm.racadm_path
            )
            if 'System Information' in simple_rac.getsysinfo:
                return True
        return False



from mercury_agent.configuration import get_configuration
from mercury.common.clients.rpc.backend import BackEndClient


# Private
__backend_client = None


def get_backend_client():
    # TODO: Trying this out, 0mq says it is ok
    global __backend_client
    if not __backend_client:
        __backend_client = BackEndClient(
            get_configuration().agent.remote.backend_url)
    return __backend_client



import os
import time
import logging

from subprocess import Popen, PIPE

log = logging.getLogger(__name__)



class LSIRaidException(Exception):
    """ Raised exclusively by Megacli classes """
    pass


class LSIRaidTimeoutException(LSIRaidException, OSError):
    pass


def clear_lsi(lsi_raid, adapter=None, wait_period=10):
    # TODO: If adapter is blank, attempt to only clear controllers
    # we know about to potentially not blow away controllers Core doesn't
    # want us to!
    wait_period = wait_period if wait_period >= 1 else 1

    log.info('Clearing config')
    lsi_raid.clear_config(adapter)

    log.info('Clearing foreign configs, if present')
    lsi_raid.clear_foreign_configs(adapter)
    time.sleep(wait_period)

    log.info('Refreshing RAID Controller')
    lsi_raid.refresh()


class LSIRaid(object):
    adapter = None

    def __init__(self, megacli_bin='/usr/sbin/megacli', adapter=0, use_sudo=False):

        self.adapter = adapter
        self.megacli_bin = megacli_bin
        if not os.path.isfile(self.megacli_bin):
            raise LSIRaidException('megacli binary is missing')
        self.use_sudo = use_sudo
        self.pdisks, self.vdisks, self.enclosure, self.raw_adapter_info, \
            self.clear = None, None, None, None, None
        self.refresh()

    def refresh(self):
        self.pdisks = self.get_physical_disks(self.adapter)
        self.vdisks = self.get_virtual_disks(self.adapter)
        self.enclosure = self.get_enclosure_id(self.adapter)
        self.raw_adapter_info = self.get_adapter_raw()

        self.clear = not self.count_virtual_disks()

    def megacli(self, args, bufsize=1048567, timeout=300):
        """
        Execute a MegaCLI system command with a timeout of 300 seconds approx.
        """
        timeout = timeout if timeout >= 1 else 1
        cmd = [self.megacli_bin] + args.split()
        if self.use_sudo:
            cmd = ['sudo'] + cmd
        cmd += ['-NoLog']
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, bufsize=bufsize)
        for _ in range(1, 10 * timeout):
            if p.poll() is not None:
                out, err = p.communicate()
                ret = p.returncode
                return out, err, ret
            time.sleep(0.1)
        raise LSIRaidTimeoutException('Command took too long to execute: '
                                      '{} {}'.format(self.megacli_bin, args))

    def get_adapter_raw(self, adapter='all'):
        out, err, ret = self.megacli('-AdpAllInfo -a{0}'.format(
            adapter or self.adapter))
        if ret:
            raise LSIRaidException(
                ('Could not get adapter info - {0}: '
                 '{1} {2}').format(ret, out, err))
        return out.splitlines()

    def clear_config(self, adapter=None):
        _adapter = 'all' if adapter is None else adapter
        return self.megacli('-CfgClr -a{0}'.format(_adapter))

    def init_logical_drives(self, adapter=None):
        _adapter = 'all' if adapter is None else adapter
        return self.megacli('-LDInit -Start -LALL -a{0}'.format(_adapter))

    def clear_foreign_configs(self, adapter=None):
        _adapter = 0 if adapter is None else adapter
        return self.megacli('-CfgForeign -Clear -a{0}'.format(_adapter))

    def count_physical_disks(self):
        out, err, ret = self.megacli('-PDGetNum -a{0}'.format(self.adapter))
        for line in out.splitlines():
            if 'Adapter' in line:
                return int(line.split(':')[1].strip())
        return 0

    def count_virtual_disks(self):
        out, err, ret = self.megacli('-LDGetNum -a{0}'.format(self.adapter))
        for line in out.splitlines():
            if 'Adapter' in line:
                return int(line.split(':')[1].strip())
        return 0

    def create_array(self, raid_level, drives):
        out = '-CfgLDAdd -r' + str(raid_level)
        array_list = ['{}:{}'.format(self.enclosure, x) for x in drives]
        array_out = ' [{}]'.format(','.join(array_list))
        out = '{} {} WB RA Cached -a{}'.format(out, array_out, self.adapter)
        sout, err, ret = self.megacli(out)
        if ret:
            return dict(error=True, command=out, stdout=sout, stderr=err)
        return dict(error=False, command=out, stdout=sout, stderr=err)

    def get_enclosure_id(self, adapter=0):
        out, err, ret = self.megacli('-EncInfo -a{0}'.format(adapter))
        for line in out.splitlines():
            if 'Device ID' in line:
                try:
                    return int(line.split(':')[1].strip())
                except ValueError:
                    continue
        # this is a kludge, when no enclosure is found
        return ""

    def get_physical_disks(self, adapter=0):
        disks = list()
        count = 0
        num = self.count_physical_disks()
        if not num:
            return dict()

        out, err, ret = self.megacli('-PdList -a{0}'.format(adapter))
        options = {
            'Device Id': 'device_id',
            'Enclosure Device ID': 'enclosure',
            'Slot Number': 'slot_number',
            'Media Error Count': 'media_errors',
            'Other Error Count': 'other_errors',
            'Predictive Failure Count': 'predictive_errors',
        }

        def raise_err(key):
            if key == 'Enclosure Device ID':
                enc_out, enc_err, _ = self.megacli(
                    '-EncInfo -a{}'.format(adapter))
                raise LSIRaidException(
                    "Could not read from controller, please make sure it is "
                    "seated correctly: {} {}".format(enc_out, enc_err))
            raise LSIRaidException("Error reading raid information for "
                                   "'{}' from megacli".format(key))
        count = 1
        for x in range(0, num):
            disk = dict()
            for line in out.splitlines()[count:]:
                count += 1
                for key, value in options.items():
                    if key in line:
                        try:
                            disk[value] = int(line.split(':')[1].strip())
                        except (ValueError, IndexError):
                            raise_err(key)
                        else:
                            break
                else:
                    if 'Raw Size' in line:
                        disk['size'] = float(line.split(
                            ':')[1].split()[0].strip())
                    elif 'Firmware state' in line:
                        disk['state'] = line.split(':', 1)[1]
                        break

            disks.append(disk)

        disks.sort(key=lambda d: '%05d:%05d' % (d['device_id'],
                                                d['slot_number']))
        return disks

    def get_virtual_disk_members(self, vid, num):
        out, err, ret = self.megacli('-LDPDInfo -a{0}'.format(self.adapter))
        count = 0
        pos = 0
        pds = list()
        for line in out.splitlines():
            if 'Virtual Drive:' in line:
                if int(line.split()[2].strip()) == vid:
                    pos = out.splitlines().index(line)
                    break

        for line in out.splitlines()[pos:]:
            if 'Device Id' in line:
                pds.append(int(
                    line.split(':')[1].strip()))
                count += 1
            if count == num:
                break
        return pds

    def get_virtual_disks(self, adapter=0):
        vdisks = list()
        count = 0
        num = self.count_virtual_disks()
        if not num:
            return dict()

        out, err, ret = self.megacli('-LdInfo -Lall -a{0}'.format(adapter))
        for x in range(0, num):
            vd = dict()
            for line in out.splitlines()[count:]:
                if 'Virtual Drive:' in line:
                    vd['id'] = int(line.split(':')[1].split()[0].strip())
                    count += 1
                    continue

                if 'RAID Level' in line:
                    primary = int(line.split(':')[1].split(',')[0].split('-')[1])
                    secondary = int(line.split(':')[1].split(',')[1].split('-')[1])
                    vd['raid_level'] = (primary, secondary)
                    count += 1
                    continue

                if 'Number Of Drives' in line:
                    vd['drive_count'] = int(line.split(':')[1].split()[0].strip())
                    count += 1
                    continue

                if 'Span Depth' in line:
                    vd['span_depth'] = int(line.split(':')[1].split()[0].strip())
                    count += 1
                    break
                count += 1

            phys_ids = self.get_virtual_disk_members(
                        vd.get('id'),
                        vd.get('drive_count'))
            vd['phys_ids'] = phys_ids
            vdisks.append(vd)

        return vdisks



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury_agent.inspector.inspectors import inspector
from mercury_agent.inspector.hwlib import lspci


@inspector.expose('pci')
def pci_inspector():
    _pci = lspci.parse_nnvmmk()
    return _pci


if __name__ == '__main__':
    from pprint import pprint

    pprint(pci_inspector())




import mock
import os

from mercury.common.helpers.cli import CLIResult
from mercury_agent.hardware.raid.interfaces.megaraid import storcli

from ..base import MercuryAgentUnitTest


class StorcliTest(MercuryAgentUnitTest):
    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_run(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        s = storcli.Storcli()
        assert s.run('') == ''

        mock_cli.run.return_value = CLIResult('Error', '', 1)

        self.assertRaises(storcli.StorcliException, s.run, *('',))

        mock_cli.find_in_path.return_value = None

        self.assertRaises(storcli.StorcliException, storcli.Storcli)

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_run_json(self, mock_cli):
        with open(os.path.join(os.path.dirname(__file__),
                               '../resources/storcli.json')) as fp:
            json_data = fp.read()

        mock_cli.run.return_value = CLIResult(json_data, '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        s = storcli.Storcli()
        assert isinstance(s.run_json('/call show all'), dict)

        mock_cli.run.return_value = CLIResult('not_valid_json', '', 0)

        self.assertRaises(storcli.StorcliException, s.run_json, *('',))

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_controllers(self, mock_cli):
        with open(os.path.join(os.path.dirname(__file__),
                               '../resources/storcli.json')) as fp:
            json_data = fp.read()

        mock_cli.run.return_value = CLIResult(json_data, '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        s = storcli.Storcli()
        assert isinstance(s.controllers, list)

        mock_cli.run.return_value = CLIResult('{}', '', 0)

        def _wrapper():
            return s.controllers

        self.assertRaises(storcli.StorcliException, _wrapper)

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_delete(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'
        s = storcli.Storcli()
        assert s.delete(0) == ''

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_add(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'

        s = storcli.Storcli()
        assert s.add(0, 'r0', '32:0-1') == ''
        assert s.add(0, 'r10', '32:0-3', pdperarray=2) == ''

        self.assertRaises(storcli.StorcliException, s.add,
                          *(0, 'r10', '32:0-3'))

        with open(os.path.join(os.path.dirname(__file__),
                               '../resources/storcli_err.txt')) as fp:
            error_data = fp.read()

        mock_cli.run.return_value = CLIResult(error_data, '', 0)

        self.assertRaises(storcli.StorcliException, s.add, *(0, '', ''))

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_get_enclosures(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'

        s = storcli.Storcli()
        s.run_json = mock.Mock()
        s.run_json.return_value = {'Controllers': [{'Response Data': 'cheese'}]}

        self.assertEqual(s.get_enclosures(), ['cheese'])

        s.run_json.assert_called_with('/call/eall show all')

        s.get_enclosures(99)
        s.run_json.assert_called_with('/c99/eall show all')

        s.run_json.return_value = {}

        self.assertRaises(storcli.StorcliException, s.get_enclosures)

    def test_bad_command_status(self):
        self.assertRaises(storcli.StorcliException,
                          storcli.Storcli.check_command_status,
                          *({'Command Status': {
                              'Status': 'Nope',
                              'Description': 'Problems',
                              'Detailed Status': ['Things???']
                          }},))

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_get_disk_group(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'

        s = storcli.Storcli()
        s.run_json = mock.Mock()
        s.run_json.return_value = {
            'Controllers': [{
                'Command Status': {
                    'Status': 'Success'
                },
                'Response Data': {
                    'Response Data': {}
                }}]}

        s.get_disk_group()

        s.run_json.assert_called_with('/call/dall show all')

        s.get_disk_group(99, 100)
        s.run_json.assert_called_with('/c99/d100 show all')

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.storcli.cli')
    def test_add_hotspare(self, mock_cli):
        mock_cli.run.return_value = CLIResult('', '', 0)
        mock_cli.find_in_path.return_value = '/sbin/storcli64'

        s = storcli.Storcli()
        s.run = mock.Mock()

        s.add_hotspare(0, 32, 10)

        s.run.assert_called_with('/c0/e32/s10 add hotsparedrive')

        s.add_hotspare(0, 32, 10, [0, 1, 2])

        s.run.assert_called_with('/c0/e32/s10 add hotsparedrive DGs=0,1,2')



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury_agent.inspector.inspectors import inspectors, late_inspectors
from mercury_agent.hardware.drivers import registered_drivers, set_driver_cache
from mercury.common.mercury_id import generate_mercury_id
from mercury.common.exceptions import fancy_traceback_short, parse_exception

log = logging.getLogger(__name__)

# Global storage for device_info it is mostly read only and only overwritten during
# inspector runs

global_device_info = {}


def _collect():
    _c = dict()
    for inspector, f in inspectors:
        _c[inspector] = f()
    return _c


def inspect():
    """
    Runs inspectors and associates collection with a mercury_id
    :return:
    """
    collected = _collect()
    dmi = collected.get('dmi') or {}
    interfaces = collected.get('interfaces') or {}

    collected['mercury_id'] = generate_mercury_id(dmi, interfaces)

    # populate_drivers

    for driver in registered_drivers:
        _wants = driver['class'].wants

        # noinspection PyBroadException
        try:
            devices = driver['class'].probe(
                _wants and collected[_wants] or collected)
        except Exception:
            # probe is implemented in each driver and is not wrapped
            # handle probe errors gracefully and soldier on
            log.error(fancy_traceback_short(
                parse_exception(),
                preamble='Probe function failed for driver {}'.format(
                    driver['name']
                )))
            continue
        if devices:
            set_driver_cache(driver, devices)

    # TODO: Sort RAID drivers based on devices

    for inspector, f in late_inspectors:
        collected[inspector] = f(collected)

    global global_device_info
    global_device_info.update(**collected)

    return global_device_info


if __name__ == '__main__':
    from pprint import pprint
    pprint(inspect())



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Unit tests for mercury_agent.capabilities"""

import mock

from mercury_agent import capabilities


def test_add_capability():
    """Test add_capability()"""
    capabilities.add_capability('c_entry',
                                'c_name',
                                'Capability Description')

    assert 'c_name' in capabilities.runtime_capabilities

    fake_capability = capabilities.runtime_capabilities['c_name']
    assert fake_capability['name'] == 'c_name'
    assert fake_capability['entry'] == 'c_entry'
    assert fake_capability['description'] == 'Capability Description'

    del capabilities.runtime_capabilities['c_name']


@mock.patch('mercury_agent.capabilities.add_capability')
def test_capability(mock_add_capability):
    """Test @capability() decorator"""
    @capabilities.capability('c_name', 'Capability Description', num_args=1)
    def c_entry(x):
        """Return input."""
        return x

    c_entry('foo')

    mock_add_capability.assert_called_once_with(c_entry,
                                                'c_name',
                                                'Capability Description',
                                                num_args=1,
                                                doc="Return input.",
                                                serial=False,
                                                kwarg_names=None,
                                                no_return=False,
                                                dependency_callback=None,
                                                timeout=1800,
                                                task_id_kwargs=False)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import subprocess


from mercury_agent.capabilities import capability
from mercury.common.helpers.cli import run
from mercury.common.helpers.util import download_file


log = logging.getLogger(__name__)


@capability('echo', 'Echo something to the console', num_args=1)
def echo(message):
    """
    Echo the dolphin
    :param message: message to Echo
    :return: None
    """
    log.info('Echo: %s' % message)
    print(message)
    return message


@capability('run', 'Run an arbitrary command', num_args=1)
def runner(command, _input=''):
    """
    Run a shell command
    :param command: The shell command to use
    :param _input: Optional data to pass to stdin
    :return:
    """
    log.info('Running: %s' % command)

    r = run(command, ignore_error=True, raise_exception=False, _input=_input)
    return {
        'stdout': r.stdout,
        'stderr': r.stderr,
        'returncode': r.returncode
    }


@capability('run_async', 'Run a command in the background', num_args=1)
def runner_async(command, shell=True):
    """
    
    :param command: 
    :param shell:
    :return: 
    """
    subprocess.Popen('{}'.format(command), shell=shell)


@capability('kexec', 'kexec into kernel at supplied location', kwarg_names=['kernel', 'initrd', 'options'],
            no_return=True, serial=True)
def kexec(kernel='', initrd='', options=None, kernel_type='bzImage'):
    """
    Kexec into a kernel
    """
    options = options or []
    command = 'kexec --type {kernel_type} --initrd={initrd} --append="{options}" {kernel}'.format(
        kernel_type=kernel_type,
        initrd=initrd,
        options=' '.join(options),
        kernel=kernel
    )

    # TODO: implement workflow that allows an agent to un-register itself
    # Sleep a little bit to allow the command to return
    log.info('Running Kexec: {}'.format(command))
    runner_async('sleep 5;' + command)


@capability('reload', 'kexec into current preboot kernel, re-downloading the root file system',
            num_args=2)
def reload(kernel_url, initrd_url):
    """
    Reload the environment
    """
    # This should look into the configuration to find the location
    # of the kernel/initrd images and download them

    kernel_file = '/tmp/vmlinuz'
    initrd_file = '/tmp/initrd'

    log.info('Downloading: {}'.format(kernel_url))
    download_file(kernel_url, kernel_file)

    log.info('Downloading: {}'.format(initrd_url))
    download_file(initrd_url, initrd_file)

    with open('/proc/cmdline') as fp:
        options = fp.readline().split()

    kexec(kernel=kernel_file, initrd=initrd_file, options=options)






import mock

from mercury.common.helpers.cli import CLIResult
from mercury_agent.procedures.stress import Stress

from tests.unit import base


class TestStress(base.MercuryAgentUnitTest):

    @mock.patch('mercury_agent.procedures.stress.os')
    @mock.patch('mercury_agent.procedures.stress.cli')
    def test_get_system_memory(self, mock_cli, mock_os):
        s = Stress()

        mock_cli.run = mock.Mock(return_value=CLIResult(
            'MemFree:          33554432 kB', '', 0))

        self.assertEqual(s.get_system_memory(), '32G')

    @mock.patch('mercury_agent.procedures.stress.os')
    def test_kill_all(self, mock_os):
        Stress()
        mock_os.system()

        mock_os.system.assert_called_once_with()
        mock_os.return_value = CLIResult('', '', 0)

    @mock.patch('mercury_agent.procedures.stress.os')
    @mock.patch('mercury_agent.procedures.stress.cli')
    def test_memory_stress(self, mock_cli, mock_os):
        s = Stress()

        cmd = '{0} --vm 1 --vm-bytes {1} --vm-hang 5 -t {2}'.format(
            '/usr/bin/stress', '1024M', 5)
        mock_cli.run = mock.Mock()
        s.memory_stress(timeout=5)

        mock_cli.run.assert_called_with(cmd)

    @mock.patch('mercury_agent.procedures.stress.os')
    @mock.patch('mercury_agent.procedures.stress.cli')
    def test_cpu_stress(self, mock_os, mock_cli):
        s = Stress()
        s.cpu_stress(timeout=5)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury.common.exceptions import MercuryUserError

from mercury_agent.capabilities import capability
from mercury_agent.backend_client import get_backend_client
from mercury_agent.hardware.drivers.drivers import driver_class_cache
from mercury_agent.inspector.inspect import global_device_info
from mercury_agent.inspector.inspectors.raid import raid_inspector


log = logging.getLogger(__name__)


def get_hp_raid_driver():
    hp_raid_driver = driver_class_cache.get('hpssa')
    if not hp_raid_driver:
        # Once dependent capabilities are added, this will no longer be necessary
        log.error('Attempt to use platform specific procedure without supporting driver')
        raise MercuryUserError('Required driver, hpssa, is not loaded. Check platform')
    return hp_raid_driver


def has_hp_raid_driver():
    return bool(driver_class_cache.get('hpssa'))


def update_inventory():
    backend_client = get_backend_client()
    raid_info = raid_inspector(global_device_info)
    mercury_id = global_device_info['mercury_id']

    log.debug('RAID configuration changed, updating inventory')
    backend_client.update(mercury_id, {'raid': raid_info})


def update_on_change(f):
    def wrapped_f(*args, **kwargs):
        result = f(*args, **kwargs)
        update_inventory()
        return result
    wrapped_f.__name__ = f.__name__
    wrapped_f.__doc__ = f.__doc__
    return wrapped_f


@capability('hpssa_create_array',
            description='Create array on an HP SmartArray Controller',
            kwarg_names=['slot', 'selection', 'raid'],
            serial=True,
            dependency_callback=has_hp_raid_driver,
            timeout=120
            )
@update_on_change
def hpssa_create_array(slot, selection, raid, array_letter=None, array_type='ld', size='max',
                       stripe_size='default', write_policy='writeback', sectors=32, caching=True,
                       data_ld=None, parity_init_method='default'):
    """
    Create an array

    :param slot: Slot ID of the adapter you are targeting
    :param selection: all, allunassigned, Port:Box:Bay,...  , 1I:1:1-1I:1:6
    :param raid: 0, 1, 5, 6, 1+0, 1+0asm, 50, 60
    :param array_letter: Optional array ID, Any unused, A-Z
    :param array_type: ld, ldcache, arrayr0
    :param size: size in MB, min, max, maxmbr
    :param stripe_size: 2**3-10 (8-1024), default
    :param write_policy: writeback, writethrough
    :param sectors: 32, 64
    :param caching: True | False
    :param data_ld: ld ID, required if array_type == ldcache
    :param parity_init_method: default

    :return type dict: stdout, stderr, returncode
    """

    hp_raid_driver = get_hp_raid_driver()

    log.info('Creating HPSSA Array: {0} {1} {2}'.format(slot, str(selection), raid))

    result = hp_raid_driver.handler.hpssa.create(
        slot,
        selection,
        raid,
        array_letter=array_letter,
        array_type=array_type,
        size=size,
        stripe_size=stripe_size,
        write_policy=write_policy,
        sectors=sectors,
        caching=caching,
        data_ld=data_ld,
        parity_init_method=parity_init_method
    )

    if result.returncode:
        log.error('Error creating array')
        raise MemoryError('Failed to create array: stdout={0}\nstderr={1}'.format(result,
                                                                                  result.stderr))

    return {'stdout': result, 'stderr': result.stderr, 'returncode': result.returncode}


@capability('hpssa_delete_ld', description='Delete a logical drive on a given controller',
            kwarg_names=['slot', 'logical_drive'], serial=True,
            dependency_callback=has_hp_raid_driver, timeout=120)
@update_on_change
def hpssa_delete_ld(slot, logical_drive):
    """
    Delete a logical drive
    :param slot: Adapter slot
    :param logical_drive: Logical drive id
    :return dict: stdout, stderr, returncode
    """
    hp_raid_driver = get_hp_raid_driver()
    result = hp_raid_driver.handler.hpssa.hpssa.delete_logical_drive(slot, logical_drive)

    return {'stdout': result, 'stderr': result.stderr, 'returncode': result.returncode}


@capability('hpssa_clear_configuration', description='Delete all arrays on a given controller',
            kwarg_names=['slot'], serial=True, dependency_callback=has_hp_raid_driver,
            timeout=120)
@update_on_change
def hpssa_clear_configuration(slot):
    """
    Delete all arrays on a given controller
    :param slot: Adapter slot
    :return dict:  stdout, stderr, returncode
    """
    hp_raid_driver = get_hp_raid_driver()
    result = hp_raid_driver.handler.hpssa.delete_all_logical_drives(slot)

    return {'stdout': result, 'stderr': result.stderr, 'returncode': result.returncode}


@capability('hpssa_clear_configuration_all_controllers',
            description='Delete all configurations from all RAID controllers',
            serial=True,
            dependency_callback=has_hp_raid_driver,
            timeout=120)
@update_on_change
def hpssa_clear_configurations_all_controllers():
    """
    Nuke it from orbit. It's the only way to be sure
    :return dict: Indexed by adapter slot
    """
    hp_raid_driver = get_hp_raid_driver()
    return hp_raid_driver.handler.hpssa.clear_configuration()


@capability('hpssa_add_spares',
            description='Add spare drives to one, many, or all arrays',
            serial=True,
            dependency_callback=has_hp_raid_driver,
            kwarg_names=['slot', 'selection'],
            timeout=120)
@update_on_change
def hpssa_add_spares(slot, selection, array_letters=None):
    """

    :param slot: Adapter slot to target
    :param selection: all, allunassigned, Port:Box:Bay,...  , 1I:1:1-1I:1:6
    :param array_letters: None for a global hotspare, otherwise a list of
        array letters that the spares should be dedicated or shared between
    :return:
    """
    results = get_hp_raid_driver().handler.hpssa.add_spares(
        slot, selection, array_letters)

    return [{'stdout': result,
             'stderr': result.stderr,
             'returncode': result.returncode} for result in results]



import logging

log = logging.getLogger(__name__)


registered_drivers = list()
driver_class_cache = dict()


class DriverBase(object):
    """
    Drivers
    -------

    """
    name = ''
    driver_type = ''
    _handler = None
    wants = ''

    def __init__(self, devices):
        """
        :param devices: Devices handled by the driver
        """
        self.devices = devices

        if self._handler:
            self.handler = self._handler()
        else:
            self.handler = None

    @classmethod
    def probe(cls, context_data):
        raise NotImplementedError

    def inspect(self):
        raise NotImplementedError


class PCIDriverBase(DriverBase):
    @classmethod
    def probe(cls, pci_data):
        """
        Implementations should return a sorted list of PCI devices supported by the handler
        :param pci_data:
        :return:
        """
        raise NotImplementedError

    @classmethod
    def check(cls, pci_device):
        """
        Implementations should check if a PCI device is handled by the driver
        :param pci_device: device_info.pci element
        :return:
        """
        raise NotImplemented

    def inspect(self):
        raise NotImplemented


def driver():
    def decorator(cls):
        log.info('Registering driver: %s' % cls.__name__)
        registered_drivers.append({
            'name': cls.name,
            'class': cls,
            'driver_type': cls.driver_type,
            'wants': cls.wants
        })

        return cls
    return decorator


def get_subsystem_drivers(subsystem):
    drivers = []
    for d in driver_class_cache:
        if driver_class_cache[d].driver_type == subsystem:
            drivers.append(driver_class_cache[d])
    return drivers


def set_driver_cache(d, devices):
    """
    Instantiate the driver and store it for later
    :param d: The driver to instantiate
    :param devices: return of the probe
    :return:
    """
    if not driver_class_cache.get(d['name']):
        log.info('Initializing d: %s' % d['name'])
        driver_class_cache[d['name']] = d['class'](devices)



"""
Drive erase methods for targeting os level block devices
"""

import logging
import os
import random
import time
import uuid
from datetime import timedelta

from press.helpers import parted

from mercury.common.helpers import cli
from mercury.common.exceptions import MercuryGeneralException, MercuryCritical

log = logging.getLogger(__name__)


def get_block_device_size(device):
    return parted.PartedInterface(device).get_size()


def open_target_sync(target):
    """ Get file descriptor with synchronous writes """
    return os.open(target, os.O_WRONLY | os.O_DSYNC)


def write_random_signatures(target, target_size_bytes, number):
    signature_map = [
        (uuid.uuid4().bytes, random.randint(0, int(
            (target_size_bytes-16)/16))*16)
        for _ in range(number)]

    target_fd = open_target_sync(target)
    try:
        for sig, offset in signature_map:
            os.pwrite(target_fd, sig, offset)
    finally:
        os.close(target_fd)

    return signature_map


def validate_erased(target, signature_map):
    target_fd = os.open(target, os.O_RDONLY)
    try:
        for sig, offset in signature_map:
            if os.pread(target_fd, 16, offset) == sig:
                return False
    finally:
        os.close(target_fd)
    return True


def buffered_zero_fill(target, target_size_bytes, buffer_size_bytes=4194304):
    """ Native python buffered zero fill implementation """
    # less memory efficient, but fastest
    zero_buffer = b'\x00' * buffer_size_bytes
    target_fd = open_target_sync(target)
    try:
        for _ in range(0, target_size_bytes, buffer_size_bytes):
            os.write(target_fd, zero_buffer)
    finally:
        os.close(target_fd)


def ssd_swap_discard(target):
    """ Use swapspace discard support to trigger TRIM across the entire block
    device """
    def _run(command):
        try:
            return cli.run(command, raise_exception=True)
        except cli.CLIException as e:
            raise MercuryCritical(
                'CLI error during drive erase; target: {}, error: {}'.format(
                    target, e))

    log.info('Creating swapspace on %s', target)

    _run('mkswap --force {}'.format(target))

    log.info('Activating swap space with the discard option')
    _run('swapon --discard=once {}'.format(target))

    log.info('Deactivating swap')
    _run('swapoff {}'.format(target))

    log.info('Removing gpt/mbr/file system signatures')
    _run('wipefs -o 0xff6 {}'.format(target))


def erase(target, method='zero', buffer_size_bytes=4194304,
          verification_blocks=4):
    """ Erase block device using the provided method
    :param target:
    :param method:
    :param buffer_size_bytes:
    :param verification_blocks:
    :return: statistics
    """
    start = time.time()
    target_size = get_block_device_size(target)

    log.info('Writing %s verification signatures to %s',
             verification_blocks, target)
    signature_map = write_random_signatures(
        target, target_size, verification_blocks)

    log.info('Beginning %s operation on %s [size: %s]', method, target,
             target_size)
    if method == 'zero':
        buffered_zero_fill(target, target_size, buffer_size_bytes)
    elif method == 'ssd_trim':
        ssd_swap_discard(target)
    else:
        raise MercuryGeneralException(
            f'Unsupported erase method specified: {method}')

    if not validate_erased(target, signature_map):
        raise MercuryCritical('Could not validate erase procedure, '
                              'Signatures are intact! {}'.format(signature_map))
    completed_at = time.time()
    time_taken = completed_at - start
    speed_stat = (target_size / time_taken / 2**20)  # MB/s
    log.info('Wipe operation completed in %s [%s MB/s]', timedelta(
        seconds=round(completed_at - start)), speed_stat)

    return {'start_time': start, 'completed_at': completed_at,
            'target_size': target_size, 'target': target,
            'speed': f'{speed_stat} MB/s'}




from mercury_agent.inspector.inspectors import expose
from mercury_agent.inspector.hwlib.meminfo import parse_meminfo


@expose('mem')
def memory_inspector():
    return parse_meminfo()


if __name__ == '__main__':
    import pprint

    pprint.pprint(memory_inspector())



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import netifaces

from mercury_agent.inspector.inspectors import inspector

from mercury_agent.inspector.hwlib import biosdevname
from mercury_agent.inspector.hwlib.sysfs import NetClass
from mercury_agent.inspector.hwlib.udev import UDevHelper
from mercury_agent.inspector.hwlib import network_interfaces

log = logging.getLogger(__name__)


def get_udev_interface_by_name(interfaces, name):
    for interface in interfaces:
        if interface.get('INTERFACE') == name:
            return interface


def index_gateways(gateways):
    _d = {}
    for gateway in gateways:
        if gateway == 'default':
            continue
        log.debug(gateway)
        _d[gateway[1]] = list()
        _d[gateway[1]].append(dict(list(zip(['gateway_ip', 'interface', 'default'], gateway))))
    return _d


@inspector.expose('interfaces')
def interface_inspector():
    """
    Doc this by providing an example interfaces data structure
    :return: interface data structure
    """
    i = []
    uh = UDevHelper()
    udev_interfaces = uh.get_network_devices()
    interfaces = network_interfaces.list_interfaces()
    gateways = netifaces.gateways()
    log.debug(gateways)

    for interface in interfaces:
        log.debug('Inspecting: {}'.format(interface))
        ndi = NetClass(interface)
        _iface = dict()
        _iface['devname'] = interface
        address = ndi.address
        if not address:
            continue
        _iface['address'] = address
        _iface['carrier'] = ndi.carrier
        log.debug('Interface {} is {}'.format(interface, ndi.carrier and 'up' or 'down'))
        _iface['dev_port'] = ndi.dev_port
        _iface['duplex'] = ndi.duplex
        _iface['speed'] = ndi.speed
        _iface['predictable_names'] = {}
        try:
            _iface['predictable_names']['biosdevname'] = biosdevname.get_name(interface)
        except OSError:
            _iface['predictable_names']['biosdevname'] = None

        udev_interface = get_udev_interface_by_name(udev_interfaces, interface) or dict()
        _iface['predictable_names']['systemd_udev'] = udev_interface.get('ID_NET_NAME_PATH')
        _iface['predictable_names']['systemd_onboard'] = udev_interface.get('ID_NET_NAME_ONBOARD')
        _iface['predictable_names']['systemd_mac'] = udev_interface.get('ID_NET_NAME_MAC')
        _iface['predictable_names']['systemd_slot'] = udev_interface.get('ID_NET_NAME_SLOT')

        if not udev_interface:
            udev_parent = dict()
        else:
            udev_parent = udev_interface.parent or dict()
        _iface['pci_slot'] = udev_parent.get('PCI_SLOT_NAME')
        _iface['model_name'] = udev_parent.get('ID_MODEL_FROM_DATABASE')
        _iface['vendor_name'] = udev_parent.get('ID_VENDOR_FROM_DATABASE')
        _iface['pci_class'] = udev_parent.get('PCI_CLASS')
        _iface['pci_id'] = udev_parent.get('PCI_ID')
        _iface['pci_subsystem_id'] = udev_parent.get('PCI_SUBSYS_ID')
        _iface['driver'] = udev_parent.get('DRIVER')

        # provide per interface routing table
        _iface['address_info'] = network_interfaces.get_ipv4_network_info(interface)

        ipv4_gateways = gateways.get(netifaces.AF_INET, [])
        _iface['ipv4_gateways'] = index_gateways(ipv4_gateways).get(interface)

        _iface['address_info_v6'] = network_interfaces.get_ipv6_network_info(interface)

        ipv6_gateways = gateways.get(netifaces.AF_INET6, [])
        _iface['ipv6_gateways'] = index_gateways(ipv6_gateways).get(interface)

        # TODO: Add gateway information from netifaces.gateways()
        i.append(_iface)

    return i


def get_interface_by_name(interfaces, name):
    """
    Return an interface by it's devname
    :param name: interface devname
    :param interfaces: interfaces dictionary provided by interface_inspector
    :return: interface dictionary
    """
    for interface in interfaces:
        if interface['devname'] == name:
            return interface


if __name__ == '__main__':
    from pprint import pprint
    logging.basicConfig(level=logging.DEBUG)
    pprint(interface_inspector())



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Module to unit test mercury.common.misc.iproute2"""

import mock
import pytest

import mercury_agent.inspector.hwlib.iproute2 as ipr2
from tests.unit.base import MercuryAgentUnitTest


IPROUTE2_EXAMPLE_OUTPUT_LINES = [
'default via 192.168.1.1 dev br0  proto static  metric 425',
'192.168.1.0/24 dev br0  proto kernel  scope link  src 192.168.1.175  metric 425',
'192.168.122.0/24 dev virbr0  proto kernel  scope link  src 192.168.122.1 linkdown'
]

EXPECTED_EXAMPLE_ENTRIES = [
    {
        'via': '192.168.1.1',
        'dev': 'br0',
        'destination': 'default',
        'proto': 'static',
        'metric': '425'
    },
    {
        'src': '192.168.1.175',
        'destination': '192.168.1.0/24',
        'dev': 'br0',
        'proto': 'kernel',
        'scope': 'link',
        'metric': '425'
    },
    {
        'src': '192.168.122.1',
        'destination': '192.168.122.0/24',
        'dev': 'virbr0',
        'proto': 'kernel',
        'scope': 'link',
        'linkdown': True
    }
]


def construct_ipr2_with_fake_output(ipr_output, ip_path='/usr/sbin/ip'):
    """Get an IPRoute2 object with faked ip route output."""
    with mock.patch('mercury.common.helpers.cli.find_in_path') as \
            find_in_path_mock:
        find_in_path_mock.return_value = ip_path
        with mock.patch('mercury.common.helpers.cli.run') as run_mock:
            run_mock.return_value = ipr_output
            return ipr2.IPRoute2()


class MercuryHwlibIPRoute2UnitTests(MercuryAgentUnitTest):
    """Unit tests for mercury.common.misc.iproute2"""
    def test_example_output_parsing(self):
        """Test if IPRoute2 parsed the example output correctly."""
        ipr = construct_ipr2_with_fake_output(
            '\n'.join(IPROUTE2_EXAMPLE_OUTPUT_LINES))
        result_table = ipr.table

        assert isinstance(result_table, list)
        assert len(result_table) == 3

        for index in range(0, len(result_table)):
            assert result_table[index] == EXPECTED_EXAMPLE_ENTRIES[index]

    def test_modified_example_parsing(self):
        """Test another case of IPRoute2 parsing."""
        new_test_output = '\n'.join(IPROUTE2_EXAMPLE_OUTPUT_LINES[0:2])
        ipr = construct_ipr2_with_fake_output(new_test_output)
        result_table = ipr.table

        assert isinstance(result_table, list)
        assert len(result_table) == 2

        for index in range(0, 2):
            assert result_table[index] == EXPECTED_EXAMPLE_ENTRIES[index]

    def test_ipr_output_is_malformed(self):
        """Test what happens if `ip route` output is malformed."""
        test_output = 'this is not good ip route output :('
        with pytest.raises(Exception):
            construct_ipr2_with_fake_output(test_output)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Bare minimum to get system routing information
"""

from mercury.common.helpers import cli


class IPRoute2(object):
    def __init__(self, path='ip'):
        self.ip_path = cli.find_in_path(path)
        self.raw_table = self.get_table()
        self.table = []
        self.parse_table()

    def ip(self, args):
        command = '%s %s' % (self.ip_path, args)
        return cli.run(command)

    @staticmethod
    def _dzip(l):
        _d = {}
        length = len(l)
        if length % 2:
            raise Exception('The list length is ODD, cannot unzip')
        for idx in range(0, len(l), 2):
            _d[l[idx]] = l[idx+1]
        return _d

    def parse_table(self):
        singletons = ['dead', 'onlink', 'pervasive', 'offload', 'notify', 'linkdown']
        for line in self.raw_table.splitlines():
            if line:
                line = line.split()
                route = {'destination': line[0]}
                for singleton in singletons:
                    if singleton in line:
                        route[singleton] = True
                        line.remove(singleton)

                route.update(self._dzip(line[1:]))
                self.table.append(route)

    def get_table(self):
        return self.ip('route show')


if __name__ == '__main__':
    ip_route = IPRoute2()
    from pprint import pprint
    pprint(ip_route.table)



import json
import os
import random

import mock

from hpssa.hpssa import Adapter

from mercury_agent.hardware.drivers.hp_raid import (
    RAIDAbstractionException,
    SmartArrayActions,
    SmartArrayDriver
)
from ..base import MercuryAgentUnitTest


def get_adapters():
    with open(os.path.join(os.path.dirname(__file__), '../resources/adapters.json')) as fp:
        return json.load(fp)


class DummySmartArrayActions(SmartArrayActions):
    def __init__(self):
        super(SmartArrayActions, self).__init__()
        self.hpssa = mock.Mock()
        self.hpssa.adapters = [Adapter(**a) for a in get_adapters()]
        self.hpssa.create = mock.Mock(return_value=True)
        self.hpssa.delete_logical_drive = mock.Mock(return_value=True)
        self.hpssa.delete_all_logical_drives = mock.Mock(return_value=True)
        self.hpssa.add_spares = mock.Mock(return_value=True)


class DummySmartArrayDriver(SmartArrayDriver):
    _handler = DummySmartArrayActions


# noinspection PyMethodMayBeStatic
# noinspection PyMethodOverriding
class MercurySmartArrayDriverTest(MercuryAgentUnitTest):
    def setUp(self):
        super(MercurySmartArrayDriverTest, self).setUp()

    def test_probe(self):
        with open(os.path.join(os.path.dirname(__file__), '../resources/pci_data.json')) as fp:
            pci_data = json.load(fp)

        assert SmartArrayDriver.probe(pci_data)
        assert not SmartArrayDriver.probe([])

    def test_inspect(self):
        with open(os.path.join(os.path.dirname(__file__), '../resources/pci_data.json')) as fp:
            pci_data = json.load(fp)

        devices = SmartArrayDriver.probe(pci_data)
        driver = DummySmartArrayDriver(devices)
        data = driver.inspect()
        assert data


class MercurySmartArrayActionsTest(MercuryAgentUnitTest):
    def setUp(self):
        super(MercurySmartArrayActionsTest, self).setUp()

        self.dummy_actions = DummySmartArrayActions()

    @mock.patch('mercury_agent.hardware.drivers.hp_raid.HPSSA')
    @mock.patch('mercury_agent.hardware.drivers.hp_raid.get_configuration')
    def test_real_init(self, mock_hpssa, mock_get_configuration):
        SmartArrayActions()

    def test_missing_array(self):
        original = get_adapters()[0]
        modified = original.copy()
        modified['configuration']['spares'][0]['arrays'] = ['Z']  # doesn't exist

        dummy_actions = DummySmartArrayActions()

        self.assertRaises(RAIDAbstractionException,
                          dummy_actions.transform_configuration, *(modified['configuration'], ))

    def test_get_slot(self):
        assert isinstance(self.dummy_actions.get_slot(self.dummy_actions.get_adapter_info(0)), int)

        # Code (Jared) is paranoid and tests for missing vendor info
        temp_actions = DummySmartArrayActions()
        adapter_info = temp_actions.get_adapter_info(2)
        del adapter_info['vendor_info']['slot']

        self.assertRaises(RAIDAbstractionException, temp_actions.get_slot, *(adapter_info,))

    def test_get_letter_from_index(self):
        assert self.dummy_actions.get_letter_from_index(self.dummy_actions.get_adapter_info(0), 0) == 'A'
        self.assertRaises(RAIDAbstractionException, self.dummy_actions.get_letter_from_index, *(
            self.dummy_actions.get_adapter_info(0), 100))

    def test_sort_drives(self):
        drives = [
            {
                'index': 0,
                'extra': {
                    'port': '1I',
                    'box': '1',
                    'bay': '1'
                }
            },
            {
                'index': 1,
                'extra': {
                    'port': '2I',
                    'box': '1',
                    'bay': '1'
                }
            },
            {
                'index': 2,
                'extra': {
                    'port': '2I',
                    'box': '2',
                    'bay': '1'
                }
            },
            {
                'index': 3,
                'extra': {
                    'port': '3I',
                    'box': '1',
                    'bay': '1'
                }
            },
            {
                'index': 4,
                'extra': {
                    'port': '3I',
                    'box': '2',
                    'bay': '1'
                }
            },
        ]
        random.shuffle(drives)
        self.dummy_actions.sort_drives(drives)

        for idx in range(5):
            assert drives[idx]['index'] == idx

        # Free test here testing for missing status

    def test_drive_status_failure(self):
        drives = [
            {
                'port': '2I',
                'box': '2',
                'bay': '1',
                'status': '????',
                'size': 100
            }
        ]
        self.assertRaises(RAIDAbstractionException, self.dummy_actions.transform_physical_drives, *(drives,))

    def test_create(self):
        assert self.dummy_actions.create_logical_drive(0, '10', [10, 11, 12, 13])
        assert self.dummy_actions.create_logical_drive(0, '10', [10, 11, 12, 13], size='10GiB')
        assert self.dummy_actions.create_logical_drive(0, '6', array=0)

    def test_delete_logical_drive(self):
        assert self.dummy_actions.delete_logical_drive(0, 0, 0)
        self.assertRaises(RAIDAbstractionException, self.dummy_actions.delete_logical_drive, *(0, 0, 100))

    def test_clear_configuration(self):
        assert self.dummy_actions.clear_configuration(0)

    def test_add_spares(self):
        assert self.dummy_actions.add_spares(0, [10, 11])
        assert self.dummy_actions.add_spares(0, [10], [0])

    def test_get_bad_adapter_info(self):
        self.assertRaises(RAIDAbstractionException, self.dummy_actions.get_adapter_info, *(100, ))

    def test_get_adapter_info(self):
        adapter_info = self.dummy_actions.get_adapter_info(0)
        self.assertEqual(adapter_info['total_drives'], 26)
        self.assertEqual(adapter_info['total_size'], 7800000000000)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury_agent.hardware.drivers import get_subsystem_drivers
from mercury_agent.inspector.inspectors import expose_late


log = logging.getLogger(__name__)


# noinspection PyUnusedLocal
@expose_late('raid')
def raid_inspector(device_info):
    drivers = get_subsystem_drivers('raid')

    if not drivers:
        return

    _inspected = list()

    for driver in drivers:
        log.info('Running RAID inspector %s' % driver.name)

        data = driver.inspect()

        if isinstance(data, list):
            _inspected += data
        else:
            _inspected.append(data)

    return _inspected



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os

log = logging.getLogger(__name__)


def parse_cookie(path):
    try:
        with open(path) as cookie_file:
            return cookie_file.read().strip()
    except IOError as ioerror:
        log.debug('Problem reading sysfs file %s [%s]' % (path, ioerror))
        return ''


def convert_bool(data):
    if not data:
        return False

    try:
        return bool(int(data))
    except ValueError:
        log.warning('Problem coercing data to int, probably not a bool : {}'.format(data))
        return bool(data)


def append_sys(path):
    return os.path.join('/sys', path.lstrip('/'))


class SysFSBase(object):
    def __init__(self):
        self.base_path = ''

    def get_cookie(self, name):
        path = os.path.join(append_sys(self.base_path), name)
        return parse_cookie(path)


class NetClass(SysFSBase):
    class_path = 'class/net'

    def __init__(self, devname):
        super(NetClass, self).__init__()
        self.devname = devname
        self.base_path = os.path.join(self.class_path, self.devname)

    @property
    def address(self):
        return self.get_cookie('address')

    @property
    def carrier(self):
        return convert_bool(self.get_cookie('carrier'))

    @property
    def dev_port(self):
        return self.get_cookie('dev_port')

    @property
    def dev_id(self):
        return self.get_cookie('dev_id')

    @property
    def duplex(self):
        return self.get_cookie('duplex')

    @property
    def speed(self):
        return self.get_cookie('speed')

    @property
    def ifindex(self):
        return self.get_cookie('ifindex')

    @classmethod
    def list_interfaces(cls, exclude_loopback=True):
        path = append_sys(cls.class_path)
        interfaces = os.listdir(path)
        if exclude_loopback and 'lo' in interfaces:
            interfaces.remove('lo')
        return interfaces


class DMI(SysFSBase):
    """\
    Since class/dmi is unlikely to change, we'll build attributes off fs objects
    """
    class_path = 'class/dmi/id'

    def __init__(self):
        super(DMI, self).__init__()
        self.base_path = self.class_path
        self.elements = list()
        self.__build_attributes()

    def __build_attributes(self):
        ld = os.listdir(append_sys(self.base_path))
        for f in ld:
            sys_path = os.path.join(self.base_path, f)
            if os.path.isfile(append_sys(sys_path)):
                self.elements.append(f)
                self.__setattr__(f, self.get_cookie(f))

    def __getattr__(self, item):
        """\
        This is only here to shut up pycharm inspections. In version 4.5, it now does not complain about missing
        attributes if __getattr__ is defined in new style classes
        """
        raise AttributeError('Attribute lookup failed for %s' % item)

    def dump(self):
        """\
        Dump a dictionary of dmi elements as key/value pairs
        """
        _d = {}
        for element in self.elements:
            _d[element] = getattr(self, element)
        return _d




from mercury.common.exceptions import MercuryGeneralException
from mercury.common.helpers import cli

HWEVENT_SEARCH_TERM = 'Hardware event. This is not a software error.'
JOURNALCTL_COMMAND = 'journalctl -a --output cat --unit=mcelog.service ' \
                     '--no-pager'


def get_mcelog_journal_stream():
    p = cli.run(JOURNALCTL_COMMAND, raw=True)
    if p.returncode:
        raise MercuryGeneralException('Error getting mcelog')

    return p.stdout


def query_mcelog_daemon(mcelog_path='mcelog'):
    """
    Used to expose memory error counts
    :param mcelog_path:
    :return:
    """
    mcelog = cli.find_in_path(mcelog_path)
    if not mcelog:
        raise MercuryGeneralException('Could not find mcelog')

    result = cli.run(f'{mcelog} --client', raise_exception=False)
    return result.stdout


def count_logged_events():
    """
    Searches data stream and returns a count of HWEVENT_SEARCH_TERM
    :param data:
    :return:
    """
    count = 0
    for line in get_mcelog_journal_stream().readlines():
        if HWEVENT_SEARCH_TERM in line.decode('utf-8'):
            count += 1
    return count


if __name__ == '__main__':
    print(count_logged_events())



import pkg_resources

from mercury_agent.inspector.inspectors import expose


@expose('agent_info')
def agent_inspector():
    _info = {
        'agent_version':
        pkg_resources.get_distribution('mercury-agent').version,
        'mercury_version':
        pkg_resources.get_distribution('mercury-core').version,
    }

    try:
        with open('/etc/hostname') as fp:
            hostname = fp.read().strip()
    except (IOError, OSError):
        hostname = None

    _info['hostname'] = hostname
    return _info


if __name__ == '__main__':
    from pprint import pprint
    pprint(agent_inspector())



import logging
import json

from mercury.common.helpers import cli


log = logging.getLogger(__name__)


class StorcliException(Exception):
    """ Raised exclusively by Storcli classes """
    pass


class Storcli(object):
    """ Thin storcli/percli interface for querying PERC and MegaRAID adapters and creating simple arrays """

    span_arrays = [10, 50, 60]
    parity_arrays = [5, 6, 50, 60]

    def __init__(self, binary_path='storcli'):
        """ Constructor

        :param binary_path: Location (relative, in shell path, or absolute) of storcli binary
        """
        self.storecli_path = cli.find_in_path(binary_path)
        if not self.storecli_path:
            raise StorcliException('storcli binary is missing')

    def run(self, cmd, ignore_error=False):
        """ Run a storecli command

        :param cmd: A valid storcli command
        :param ignore_error: Do not raise an exception on non-zero return code
        :return: AttributeString
        """

        result = cli.run('{} {}'.format(self.storecli_path, cmd), ignore_error=ignore_error)
        if not ignore_error:
            if result.returncode:
                raise StorcliException('Command returned: {}, Error: {}'.format(
                    result.returncode, result))
        return result

    def run_json(self, cmd):
        """ Request json output from storecli and load the object it returns

        :param cmd: A valid storcli command
        :return: deserialized storcli output
        """

        try:
            loaded = json.loads(self.run(cmd + ' J'))
        except ValueError as decode_error:
            raise StorcliException('Problem processing output: {}'.format(decode_error))
        return loaded

    @property
    def controllers(self):
        """ Property used to fetch, format, and display Controller information
        :return: Controller info sans 'Command Status'
        :return type: list
        """
        output = self.run_json('/call show all')

        try:
            controllers = output['Controllers']
        except KeyError:
            raise StorcliException('Output is missing Controllers segment')

        return [c['Response Data'] for c in controllers]

    def get_enclosures(self, controller='all'):
        """ Returns enclosures in a flattened list
        :param controller: one or all controllers
        :return: List of enclosure objects
        """
        output = self.run_json('/c{}/eall show all'.format(controller))
        try:
            controllers = output['Controllers']
        except KeyError:
            raise StorcliException('Output is missing Controllers segment')

        return [c['Response Data'] for c in controllers]

    @staticmethod
    def check_command_status(data):
        if data['Command Status']['Status'] != 'Success':
            raise StorcliException('{}\n'.format(
                data['Command Status']['Description'],
                '\n'.join(data['Command Status']['Detailed Status'])))

    def get_disk_group(self, controller='all', disk_group='all'):
        """

        :param controller:
        :param disk_group:
        :return:
        """
        output = self.run_json('/c{}/d{} show all'.format(controller,
                                                          disk_group))

        controller_list = []
        for controller in output['Controllers']:
            self.check_command_status(controller)
            # Account for some crazy JSON schema
            controller_list.append(
                controller['Response Data']['Response Data']
            )

        return controller_list

    def delete(self, controller, virtual_drive='all'):
        """ Delete one or all virtual drives on a controller
        :param controller: Controller ID
        :param virtual_drive: Virtual Drive ID or 'all' (default all)
        :return:
        """
        return self.run('/c{} /v{} del force'.format(controller, virtual_drive))

    def add(self, controller, array_type, drives, size=None, pdperarray=None, pdcache=None,
            dimmerswitch=None, io_mode='direct', write_policy='wb', read_policy='ra',
            cachevd=False, stripe_size=None, spares=None, cached_bad_bbu=False, after_vd=None):
        """ Add a virtual drive

        :param controller: Controller ID
        :param array_type:  r[0|1|5|6|10|50|60]
        :param drives: Drives specified with as EncID:drive,...
        :param size: Size of a drive in MB or None for maximum
        :param pdperarray: Specifies the number of physical drives per array. The default
value is automatically chosen.(0 to 16)
        :param pdcache: Enables or disables PD cache. (on|off|default)
        :param dimmerswitch: Specifies the power-saving policy. Sets to default automatically.
 default: Logical device uses controller default power-saving policy.
 automatic (auto): Logical device power savings managed by firmware.
 none: No power-saving policy.
 maximum (max): Logical device uses maximum power savings.
 MaximumWithoutCaching (maxnocache): Logical device does not cache
 write to maximize power savings.
        :param io_mode: cached|direct
        :param write_policy:wb|rt
        :param read_policy:ra|rt
        :param cachevd: enables or disables cachecade device support
        :param stripe_size: stripe size ( the amount of data writen before moving to the next disk )
        :param spares: Numer drives allocated as hot spares
        :param cached_bad_bbu: Enable write caches even when the bbu is missing or discharged
        :param after_vd: Specify an existing VD to add this new vd behind
        :return: AttributeString of command output
        """

        command = '/c{controller} add vd type={array_type}{size}drives={drives}  ' \
                  '{write_policy} {read_policy} {io_mode} {stripe_size}{spares}' \
                  '{cache_bad_bbu}{after_vd}{pdperarray}'.format(
                    **{'controller': controller,
                       'array_type': array_type,
                       'drives': drives,
                       'size': size and ' size={} '.format(size) or ' ',
                       'write_policy': write_policy,
                       'read_policy': read_policy,
                       'io_mode': io_mode,
                       'pdperarray': pdperarray and 'pdperarray={} '.format(pdperarray) or '',
                       'pdcache': pdcache and 'pdcache={} '.format(pdcache) or '',
                       'dimmerswitch': dimmerswitch and 'dimmerswitch={} '.format(dimmerswitch) or '',
                       'cachevd': cachevd and 'cachevd ' or '',
                       'stripe_size': stripe_size and 'strip={} '.format(stripe_size) or '',
                       'spares': spares and 'spares={} '.format(spares) or '',
                       'cache_bad_bbu': cached_bad_bbu and 'cachebadbbu ' or '',
                       'after_vd': after_vd and 'aftervd={} '.format(after_vd) or ''
                       })

        for level in self.span_arrays:
            if str(level) in array_type:
                if not pdperarray:
                    raise StorcliException('Span depth (pdperarray) must be specified')
                break

        out = self.run(command)

        lines = [l.strip() for l in out.splitlines() if l.strip()]

        # command always returns zero on 'syntax' errors, so do some light
        # parsing to see if command succeeded
        for idx in range(len(lines)):
            if lines[idx].find('help -') == 0:
                error_output = '\n'.join(lines[:idx])
                raise StorcliException('Error creating ld, response:\n\n{}'.format(error_output))

        return out

    def add_hotspare(self, controller, enclosure, slot, disk_groups=None):
        """
        Set drive as a hot spare drive for the controller
        :param controller: The controller
        :param enclosure:
        :param slot:
        :param disk_groups:
        :return:
        """
        if disk_groups:
            dgs = ','.join(str(d) for d in disk_groups)
        else:
            dgs = None

        self.run('/c{}/e{}/s{} add hotsparedrive{}'.format(
            controller,
            enclosure,
            slot,
            dgs and ' DGs={}'.format(dgs) or ''
        ))

    def start_erase(self, controller, enclosure, slot, pattern='simple'):
        log.info('Scheduling erase [%s] operation on /c%s/e%s/s%s',
                 pattern, controller, enclosure, slot)
        data = self.run('/c{}/e{}/s{} start erase {} J'.format(
            controller,
            enclosure,
            slot,
            pattern
        ))['Controllers'][0]
        self.check_command_status(data)

    def stop_erase(self, controller, enclosure, slot):
        log.info('Cancelling erase operation on /c%s/e%s/s%s',
                 controller, enclosure, slot)
        data = self.run('/c{}/e{}/s{} stop erase J'.format(
            controller,
            enclosure,
            slot
        ))['Controllers'][0]
        self.check_command_status(data)
        return data['Response Data']

    def show_erase(self, controller, enclosure, slot):
        data = self.run('/c{}/e{}/s{} show erase J'.format(
            controller,
            enclosure,
            slot
        ))['Controllers'][0]
        self.check_command_status(data)
        return data['Response Data']



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import os

from mercury_agent.hardware.raid.abstraction.api import (
    RAIDAbstractionException,
    RAIDActions,
)

from ..base import MercuryAgentUnitTest


class DummyImplementation(RAIDActions):
    def __init__(self):
        super(DummyImplementation, self).__init__()

        with open(os.path.join(os.path.dirname(__file__), '../resources/dummy.json')) as fp:
            self.dummy_data = json.load(fp)

    def transform_adapter_info(self, adapter_index):
        try:
            return self.dummy_data[adapter_index]
        except IndexError:
            raise RAIDAbstractionException('...')

    def create(self, adapter, level, drives=None, size=None, array=None):
        return True

    def delete_logical_drive(self, adapter, array, ld):
        return True

    def clear_configuration(self, adapter):
        return True

    def add_spares(self, adapter, drives, arrays=None):
        return True

    @staticmethod
    def sort_drives(drives):
        drives.sort(key=lambda x: '{}-{:05}-{:05}'.format(
            x['extra']['port'], int(x['extra']['box']),  int(x['extra']['bay'])))


class MercuryRAIDAbstractionAPITest(MercuryAgentUnitTest):
    def setUp(self):
        super(MercuryRAIDAbstractionAPITest, self).setUp()
        self.dummy = DummyImplementation()
        self.abstract = RAIDActions()

    def test_raid_calc(self):
        _calc = self.dummy.raid_calculator

        tests = [
            dict(level='0', number=1, size=300, result=300),
            dict(level='1', number=2, size=300, result=300),
            dict(level='5', number=3, size=300, result=600),
            dict(level='6', number=6, size=300, result=1200),
            dict(level='10', number=4, size=300, result=600),
            dict(level='1+0', number=4, size=300, result=600),
            dict(level='50', number=6, size=300, result=1200),
            dict(level='60', number=8, size=300, result=1200)
        ]

        for test in tests:
            assert _calc(test['level'], test['number'], test['size']) == test['result']

        self.assertRaises(RAIDAbstractionException, _calc, *('20', 0, 0, 0))

    def test_raid_minimums(self):
        _min = self.dummy.raid_minimums
        tests = [
            dict(level='1', _pass=2, fail=1),
            dict(level='5', _pass=3, fail=2),
            dict(level='6', _pass=4, fail=3),
            dict(level='10', _pass=4, fail=7),
            dict(level='1+0', _pass=4, fail=3),
            dict(level='50', _pass=6, fail=5),
            dict(level='60', _pass=8, fail=7)
        ]

        for test in tests:
            # Will raise on failure
            self.assertEqual(_min(test['level'], test['_pass']), None)

        for test in tests:
            self.assertRaises(RAIDAbstractionException, _min, *(test['level'], test['fail']))

        self.assertRaises(RAIDAbstractionException, _min, *('60', 11))

    def test_create(self):
        # Create new array
        assert self.dummy.create_logical_drive(adapter=0, level='0', drives='9, 10', size='10GiB')
        assert self.dummy.create_logical_drive(adapter=0, level='0', drives='9-11', size='10%FREE')
        assert self.dummy.create_logical_drive(adapter=0, level='0', drives=9)
        assert self.dummy.create_logical_drive(adapter=0, level='0', drives=[9, 10, 11])

        # Update existing array
        assert self.dummy.create_logical_drive(adapter=0, level='0', array=0, size='10%FREE')

        test_exception_args = [
            (0, '0', '9, 10', '100TiB'),  # Size is too big
            (0, '0'),  # Neither drives or array is specified
            (0, '0', None, '10GiB', 100),  # Array index is invalid
            (1, '0', None, '1MiB', 0),  # Not enough free space
            (0, '0', None, '100TiB', 0),  # Array does not have enough free space
            (0, '0', None, '100%', 0),  # Only %FREE is supported
            (0, '0', '11-9', None, None),  # range is negative
            (0, '0', '9-XXX', None, None),  # range is nonsense
            (0, '0', '9_10', None, None),  # range needs a '-'
            (0, '0', '9-10-11', None, None),  # too many '-'
            (0, '0', 'all', None, None),  # All drives are not available
            (0, '0', 'unassigned', None, None),  # One of the unassigned drives is marked FAILED
        ]

        for args in test_exception_args:
            self.assertRaises(RAIDAbstractionException, self.dummy.create_logical_drive, *args)

    def test_abstract(self):
        # Silly tests for 'coverage'
        self.assertRaises(NotImplementedError, self.abstract.transform_adapter_info, *(0, ))
        self.assertRaises(NotImplementedError, self.abstract.create, *(0, 0))
        self.assertRaises(NotImplementedError, self.abstract.delete_logical_drive, *(0, 0, 0))
        self.assertRaises(NotImplementedError, self.abstract.clear_configuration, *(0, ))
        self.assertRaises(NotImplementedError, self.abstract.add_spares,
                          *(0, 0, None))
        self.abstract.sort_drives([0, 1, 2, 3])

    def test_get_drives(self):
        assert self.dummy.get_unassigned(0)

        self.assertRaises(RAIDAbstractionException,
                          self.dummy.get_unassigned, *(100, ))  # invalid adapter

    def test_add_index(self):
        # Tests that indexes are added to drives
        drives = self.dummy.get_all_drives(0)
        for idx in range(len(drives)):
            assert idx == drives[idx]['index']



import logging
from mercury.common.helpers import cli

from mercury_agent.configuration import get_configuration
from mercury_agent.hardware import platform_detection
from mercury_agent.hardware.oem.hp import hpasmcli
from mercury_agent.hardware.general import mcelog
from mercury_agent.inspector.inspectors import expose_late

log = logging.getLogger(__name__)


def _qc_hp(_oem_details, log_error):
    _oem_details['target'] = 'HP'
    hpasm = hpasmcli.HPASMCLI(
        get_configuration().agent.hardware.oem.hp.hpasmcli_path)
    _oem_details['server'] = hpasm.show_server()
    _oem_details['dimm'] = hpasm.show_dimm()
    _oem_details['power_supplies'] = hpasm.show_powersupply()

    for i, proc in enumerate(_oem_details['server'].get('processors', [])):
        if proc['status'] != 'Ok':
            log_error(
                'Processor {} is not in a nominal state. Reported state: {}'.format(
                    i, proc['status']))

    for dimm in _oem_details.get('dimm', []):
        if dimm['status'] != 'Ok':
            log_error(
                'DIMM module {} on CPU controller {} is not in a nominal state. '
                'Reporting state: {}'.format(
                    dimm['module_#'], dimm['processor_#'], dimm['status']))

    for i, power_supply in enumerate(_oem_details.get('power_supplies')):
        if power_supply['condition'] != 'Ok':
            log_error(
                'Power supply {} is not in a nominal state. Condition: {}'.format(
                    i, power_supply['condition']))


# noinspection PyTypeChecker
@expose_late('system_health')
def system_health_inspector(device_info):
    agent_configuration = get_configuration().agent
    _health = {
        'corrected_hardware_event_count': mcelog.count_logged_events(),
        'system_uptime': cli.run('uptime -s', ignore_error=True).strip(),
        'has_errors': False,
        'errors': []
    }

    def log_error(msg):
        log.error(msg)
        _health['has_errors'] = True
        _health['errors'].append(msg)

    if (_health['corrected_hardware_event_count']
            >= agent_configuration.hardware.mce_threshold):
        log_error('Hardware event count is {} which exceeds the threshold of {}'.format(
            _health['corrected_hardware_event_count'],
            agent_configuration.hardware.mce_threshold))

    # OEM Checks here
    _oem_details = {}
    _health['oem_details'] = _oem_details

    if platform_detection.is_hp(device_info.get('dmi', {})):
        _qc_hp(_oem_details, log_error)

    return _health


if __name__ == '__main__':
    from pprint import pprint

    pprint(system_health_inspector({}))



import unittest

import mock

from subprocess import PIPE
from mercury_agent.hardware.raid.interfaces.megaraid.megacli import (
    LSIRaid, LSIRaidException, clear_lsi
)


class TestLSIRaidInit(unittest.TestCase):

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.megacli.os.path.isfile')
    @mock.patch.object(LSIRaid, "refresh")
    def test_init(self, mock_refresh, mock_os_ispath):
        mock_os_ispath.return_value = True
        lsi_raid = LSIRaid()
        mock_refresh.assert_called_once()
        self.assertEqual(lsi_raid.adapter, 0)
        self.assertEqual(lsi_raid.megacli_bin, '/usr/sbin/megacli')

    @mock.patch.object(LSIRaid, 'count_virtual_disks')
    @mock.patch.object(LSIRaid, 'get_adapter_raw')
    @mock.patch.object(LSIRaid, 'get_enclosure_id')
    @mock.patch.object(LSIRaid, 'get_virtual_disks')
    @mock.patch.object(LSIRaid, 'get_physical_disks')
    @mock.patch.object(LSIRaid, '__init__', return_value=None)
    def test_refresh(self, mock_init, mock_get_physical_disks, mock_get_virtual_disks,
                     mock_get_enclosure_id, mock_get_adapter_raw, mock_count_virtual_disks):
        lsi_raid = LSIRaid().refresh()
        mock_get_physical_disks.assert_called_once()
        mock_get_virtual_disks.assert_called_once()
        mock_get_enclosure_id.assert_called_once()
        mock_get_adapter_raw.assert_called_once()
        mock_count_virtual_disks.assert_called_once()


class TestLSIRaid(unittest.TestCase):

    def setUp(self):
        self.patch_os = mock.patch(
            'mercury_agent.hardware.raid.interfaces.megaraid.megacli.os.path.isfile'
        )
        self.patch_os.start()
        self.patch_refresh = mock.patch.object(LSIRaid, "refresh")
        self.patch_refresh.start()
        self.patch_time = mock.patch(
            'mercury_agent.hardware.raid.interfaces.megaraid.megacli.time')
        self.patch_time.start()

    def tearDown(self):
        self.patch_os.stop()
        self.patch_refresh.stop()
        self.patch_time.stop()

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.megacli.Popen')
    def test_megacli(self, mock_popen):
        mock_communicte = mock.Mock(return_value=('output', None))
        mock_popen.return_value = mock.Mock(
            poll=mock.Mock(return_value=1),
            communicate=mock_communicte,
            returncode=0
        )
        cmd = '-AdpAllInfo -aall'
        out, err, ret = LSIRaid().megacli(cmd)
        expected_cmd = ['/usr/sbin/megacli'] + cmd.split()
        mock_popen.assert_called_once_with(
            expected_cmd + ['-NoLog'], stdout=PIPE, stderr=PIPE, bufsize=1048567
        )
        self.assertEqual(out, 'output')
        self.assertEqual(err, None)
        self.assertEqual(ret, 0)

    @mock.patch.object(LSIRaid, "megacli", return_value=('a1\na2\n', None, 0))
    def test_get_adapter_raw(self, mock_megacli):
        result = LSIRaid().get_adapter_raw()
        mock_megacli.assert_called_once_with('-AdpAllInfo -aall')
        self.assertEqual(result, ['a1', 'a2'])

    @mock.patch.object(LSIRaid, "megacli", return_value=(None, 'error', -1))
    def test_get_adapter_raw_error(self, mock_megacli):
        self.assertRaises(LSIRaidException, LSIRaid().get_adapter_raw)

    @mock.patch.object(LSIRaid, "megacli", return_value=('success', None, 0))
    def test_clear_config(self, mock_megacli):
        LSIRaid().clear_config()
        mock_megacli.assert_called_once_with('-CfgClr -aall')

    @mock.patch.object(LSIRaid, "megacli", return_value=('success', None, 0))
    def test_init_logical_drives(self, mock_megacli):
        LSIRaid().init_logical_drives()
        mock_megacli.assert_called_once_with('-LDInit -Start -LALL -aall')

    @mock.patch.object(LSIRaid, "megacli", return_value=('success', None, 0))
    def test_clear_foreign_configs(self, mock_megacli):
        LSIRaid().clear_foreign_configs()
        mock_megacli.assert_called_once_with('-CfgForeign -Clear -a0')

    @mock.patch.object(LSIRaid, "megacli", return_value=('Adapter:4', None, 0))
    def test_count_physical_disks(self, mock_megacli):
        result = LSIRaid().count_physical_disks()
        mock_megacli.assert_called_once_with('-PDGetNum -a0')
        self.assertEqual(result, 4)

    @mock.patch.object(LSIRaid, "megacli", return_value=('Adapter:4', None, 0))
    def test_count_virtual_disks(self, mock_megacli):
        result = LSIRaid().count_virtual_disks()
        mock_megacli.assert_called_once_with('-LDGetNum -a0')
        self.assertEqual(result, 4)

    @mock.patch.object(LSIRaid, "megacli", return_value=('Device ID:12345', None, 0))
    def test_get_enclosure_id(self, mock_megacli):
        result = LSIRaid().get_enclosure_id()
        mock_megacli.assert_called_once_with('-EncInfo -a0')
        self.assertEqual(result, 12345)

    @mock.patch.object(LSIRaid, "count_physical_disks", return_value=2)
    @mock.patch.object(LSIRaid, "megacli")
    def test_get_physical_disks(self, mock_megacli, mock_count_physical_disks):
        fake_list = ('Adapter #0\nDevice Id:1\nSlot Number:1\nFirmware state:active\nDevice Id:2'
                     '\nSlot Number:2\nFirmware state:active\n')
        mock_megacli.return_value = (fake_list, None, 0)
        result = LSIRaid().get_physical_disks()
        mock_count_physical_disks.assert_called_once()
        mock_megacli.assert_called_once_with('-PdList -a0')
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        for item in result:
            self.assertIsInstance(item, dict)
            self.assertIn('device_id', item)
            self.assertIn(item['device_id'], [1, 2])
            self.assertIn('slot_number', item)
            self.assertIn(item['slot_number'], [1, 2])
            self.assertIn('state', item)
            self.assertEqual(item['state'], 'active')

    @mock.patch.object(LSIRaid, "megacli")
    def test_get_virtual_disk_members(self, mock_megacli):
        fake_list = ('Virtual Drive vhd vid1\nDevice Id:12345')
        mock_megacli.return_value = (fake_list, None, 0)
        result = LSIRaid().get_virtual_disk_members('vid1', 1)
        mock_megacli.assert_called_once_with('-LDPDInfo -a0')
        self.assertEqual(result, [12345])

    @mock.patch.object(LSIRaid, "get_virtual_disk_members")
    @mock.patch.object(LSIRaid, "count_virtual_disks", return_value=1)
    @mock.patch.object(LSIRaid, "megacli")
    def test_get_virtual_disks(self, mock_megacli, mock_count_virtual_disks, mock_get_members):
        fake_list = (
            'Virtual Drive:1\n'
            'RAID Level:level-1,level-5\n'
            'Number Of Drives:1\n'
            'Span Depth:4\n'
        )
        mock_megacli.return_value = (fake_list, None, 0)
        mock_get_members.return_value = [12345]
        result = LSIRaid().get_virtual_disks()
        mock_count_virtual_disks.assert_called_once()
        mock_megacli.assert_called_once_with('-LdInfo -Lall -a0')
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], dict)
        expected_keys = ['id', 'raid_level',
                         'drive_count', 'span_depth', 'phys_ids']
        self.assertEqual(len(result[0].keys()), len(expected_keys))
        self.assertEqual(result[0]['id'], 1)
        self.assertEqual(result[0]['raid_level'], (1, 5))
        self.assertEqual(result[0]['drive_count'], 1)
        self.assertEqual(result[0]['span_depth'], 4)
        self.assertEqual(result[0]['phys_ids'], [12345])

    @mock.patch('mercury_agent.hardware.raid.interfaces.megaraid.megacli.LSIRaid')
    def test_clear_lsi(self, mock_lsi_raid):
        clear_lsi(mock_lsi_raid)
        mock_lsi_raid.clear_config.assert_called_once_with(None)
        mock_lsi_raid.clear_foreign_configs.assert_called_once_with(None)
        mock_lsi_raid.refresh.assert_called_once()




import re

from mercury.common.helpers import cli


invalid_charcaters = '.$ /\\\\'


class HPAsmException(Exception):
    pass


class HPASMCLI:
    def __init__(self, hpasmcli_path='hpasmcli'):
        """
        A class interface for the hpasmcli utility
        :param hpasmcli_path:
        """
        self.hpasmcli = cli.find_in_path(hpasmcli_path)

        if not self.hpasmcli:
            raise HPAsmException('Could not find hpasmcli binary')

    @staticmethod
    def fix_key_names(data):
        rgx = re.compile('[{}]'.format(invalid_charcaters))
        return rgx.sub('_', data.strip(invalid_charcaters).lower())

    @staticmethod
    def convert_digit(data):
        return int(data) if data.isdigit() else data

    def hpasm_run(self, command):
        result = cli.run('{} -s \'{}\''.format(self.hpasmcli, command))
        if result.returncode:
            raise HPAsmException('Error running command: {}'.format(command))
        return result

    def show_server(self):
        """
        Data is probably formatted like so:

            System        : ProLiant DL380 Gen9
            Serial No.    : TC51NR9952
            ROM version   : v2.60 (05/21/2018) P89
            UEFI Support  : Yes
            iLo present   : Yes
            Embedded NICs : 8
                NIC1 MAC: 38:63:bb:3f:4b:f4
                NIC2 MAC: 38:63:bb:3f:4b:f5
                NIC3 MAC: 38:63:bb:3f:4b:f6
                NIC4 MAC: 38:63:bb:3f:4b:f7
                NIC5 MAC: 8c:dc:d4:ad:d6:d0
                NIC6 MAC: 8c:dc:d4:ad:d6:d1
                NIC7 MAC: 68:05:ca:39:89:a0
                NIC8 MAC: 68:05:ca:39:89:a1

            Processor: 0
                Name         : Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz
                Stepping     : 2
                Speed        : 2400 MHz
                Bus          : 100 MHz
                Core         : 8
                Thread       : 16
                Socket       : 1
                Level1 Cache : 512 KBytes
                Level2 Cache : 2048 KBytes
                Level3 Cache : 20480 KBytes
                Status       : Ok

            Processor: 1
                Name         : Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz
                Stepping     : 2
                Speed        : 2400 MHz
                Bus          : 100 MHz
                Core         : 8
                Thread       : 16
                Socket       : 2
                Level1 Cache : 512 KBytes
                Level2 Cache : 2048 KBytes
                Level3 Cache : 20480 KBytes
                Status       : Ok

            Processor total  : 2

            Memory installed : 131072 MBytes
            ECC supported    : Yes
        """
        data = self.hpasm_run('SHOW SERVER')

        details = {}
        embedded_nics_context = False
        processor_context = False
        processor_index = -1
        for line in [_ for _ in data.splitlines() if _]:
            label, value = (_.strip() for _ in line.split(':', 1))
            label = self.fix_key_names(label)
            value = self.convert_digit(value)
            if embedded_nics_context or processor_context:
                if line[0] != '\t':
                    embedded_nics_context = processor_context = False

            if label == 'embedded_nics':
                details[label] = {'count': int(value), 'nics': []}
                embedded_nics_context = True
            elif label == 'processor':
                processor_context = True
                processor_index = value
                if processor_index == 0:
                    details['processors'] = [{}]
                else:
                    details['processors'].append({})
            elif embedded_nics_context:
                details['embedded_nics']['nics'].append({label: value})
            elif processor_context:
                details['processors'][processor_index][label] = value
            else:
                details[label] = value

        return details

    def show_dimm(self):
        """
        Example output:

            DIMM Configuration
            ------------------
            Processor #:                     1
            Module #:                     1
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     1
            Module #:                     4
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     1
            Module #:                     9
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     1
            Module #:                     12
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     2
            Module #:                     1
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     2
            Module #:                     4
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     2
            Module #:                     9
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

            Processor #:                     2
            Module #:                     12
            Present:                      Yes
            Form Factor:                  9h
            Memory Type:                  DDR4(1ah)
            Size:                         16384 MB
            Speed:                        2133 MHz
            Supports Lock Step:           No
            Configured for Lock Step:     No
            Status:                       Ok

        """
        data = self.hpasm_run('SHOW DIMM')

        # Knocks off the first three lines then splits on double \n
        segments = '\n'.join(data.splitlines()[3:]).split("\n\n")

        details = []
        for segment in segments:
            if segment:
                dimm_info = {}
                details.append(dimm_info)
                for line in segment.splitlines():
                    key, value = (_.strip() for _ in line.split(':', 1))
                    dimm_info[self.fix_key_names(key)] = self.convert_digit(value)
        return details

    def show_powersupply(self):
        data = self.hpasm_run('SHOW POWERSUPPLY')
        power_supplies = []
        ps_data = {}
        for line in [_ for _ in data.splitlines() if _]:
            if 'Power supply' in line:
                if ps_data:
                    power_supplies.append(ps_data)
                ps_data = {}
                continue
            key, value = (_.strip() for _ in line.split(':', 1))
            ps_data[self.fix_key_names(key)] = value
        power_supplies.append(ps_data)
        return power_supplies

    def clear_iml(self):
        return self.hpasm_run('CLEAR IML')



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


"""Classes for side channel configuration of various out of band management
controllers

Classes should provide:
    Verify networking configuration
    Create network configuration
    Create/enable default obm user
"""

import logging
import os

from lxml import etree

from mercury.common.helpers import cli

log = logging.getLogger(__name__)

DEFAULT_IPMITOOL_PATH = 'ipmitool'


def get_bmc_handler(vendor, **kwargs):
    _SELECTOR = {
        'Dell Inc.': IPMIToolDrac,
        'Quanta': DCMITool,
        'HP': IPMIToolHP
    }

    handler = _SELECTOR.get(vendor)
    if not handler:
        return
    return handler(**kwargs)


class BMCException(Exception):
    pass


class IPMIToolParsingError(Exception):
    pass


class IPMITool(object):
    def __init__(self, ipmitool_path=DEFAULT_IPMITOOL_PATH):
        self.ipmitool_path = cli.find_in_path(ipmitool_path)

        if not self.ipmitool_path:
            raise OSError("%s does not exist" % ipmitool_path)

        self.net_info = {}
        self.parse_network_info()

    @staticmethod
    def __lan_set(channel, what, value):
        return 'lan set %d %s %s' % (channel, what, value)

    @staticmethod
    def __raw_command(cmd):
        return 'raw %s' % cmd

    def run(self, args, bufsize=1048567, raise_on_error=True, ignore_error=False):
        cmd = self.ipmitool_path + ' ' + args
        return cli.run(cmd, bufsize=bufsize, raise_exception=raise_on_error, ignore_error=ignore_error)

    def parse_network_info(self, channel=1):
        log.debug('Parsing and storing IPMI LAN info')
        args = 'lan print %d' % channel
        out = self.run(args)
        for line in out.splitlines():
            label = line.split(':')[0].strip()
            data = line.split(':')[1].strip()
            if not label:
                continue
            if label == 'IP Address Source':
                self.net_info['source'] = 'Static' in data and 'static' or 'dhcp'
                continue
            if label == 'IP Address':
                self.net_info['ip_address'] = data
                continue
            if label == 'Subnet Mask':
                self.net_info['netmask'] = data
                continue
            if label == 'MAC Address':
                self.net_info['mac_address'] = ':'.join(line.split(':')[1:]).strip()
                continue
            if label == 'Default Gateway IP':
                self.net_info['gateway'] = data
                continue

    def verify_network_info(self, ip_address, netmask, gateway, source):
        if not self.net_info.get('ip_address') == ip_address:
            log.debug('IP: %s != %s' % (ip_address, self.net_info.get('ip_address')))
            return False

        if not self.net_info.get('netmask') == netmask:
            log.debug('NETMASK: %s != %s' % (netmask, self.net_info.get('netmask')))
            return False

        if 'gateway' in self.net_info:
            #  DCMITool does not expose default gateway
            if not self.net_info['gateway'] == gateway:
                log.debug('GATEWAY: %s != %s' % (gateway, self.net_info['gateway']))
                return False

        if not self.net_info.get('source') == source:
            log.debug('SOURCE: %s != %s' % (source, self.net_info.get('source')))
            return False

        return True

    def set_static(self, channel=1):
        self.run(self.__lan_set(channel, 'ipsrc', 'static'))

    def set_dhcp(self, channel=1):
        self.run(self.__lan_set(channel, 'ipsrc', 'dhcp'))

    def configure_static_network(self, ip_address, netmask, gateway, channel=1):
        self.set_static(channel)
        self.run(self.__lan_set(channel, 'ipaddr', ip_address))
        self.run(self.__lan_set(channel, 'netmask', netmask))
        self.run(self.__lan_set(channel, 'defgw', gateway))
        self.parse_network_info()

    def reset_mc_cold(self):
        self.run('mc reset cold')

    def create_superuser(self, username, password, channels=(1, 2), user_id=2):
        self.run('user set name %d %s' % (user_id, username))
        self.run('user set password %d %s' % (user_id, password))
        for channel in channels:
            self.run(
                'channel setaccess %d %d link=on ipmi=on callin=on privilege=4' % (channel,
                                                                                   user_id))
            self.run('sol payload enable %d %d' % (channel, user_id))
        self.run('user enable %d' % user_id)

    @staticmethod
    def parse_output_type1(data):
        key = None
        parsed = {}
        for line in data.splitlines():
            sline = line.split(':', 1)
            if len(sline) == 1:
                if not key:
                    raise IPMIToolParsingError('Key is not set for presumed list value')
                value = sline[0].strip()
                parsed[key].append(value)
                continue
            key = sline[0].strip().lower().replace(' ', '_')
            if not sline[1].strip():
                parsed[key] = list()
            else:
                parsed[key] = sline[1].strip()

        return parsed

    @property
    def bmc_info(self):
        log.debug('Getting bmc info')
        out = self.run('mc info', raise_on_error=False)
        if out.returncode:
            log.error('Problem getting bmc info : '
                      'std {} - err {} - return {}'.format(out, out.stderr, out.returncode))
            return None

        return self.parse_output_type1(out)


class DCMITool(IPMITool):
    def __init__(self, dcmitool_path='/usr/local/sbin/Qdcmitool'):
        self.dcmitool_path = cli.find_in_path(dcmitool_path)
        super(DCMITool, self).__init__(ipmitool_path=dcmitool_path)

    def bootstrap_obm_static(self, ip_address, netmask, gateway, username, password):
        self.configure_static_network(ip_address, netmask, gateway)
        self.create_superuser(username, password, channels=[1])
        return self.verify_network_info(ip_address, netmask, gateway, 'static')


class IPMIToolDrac(IPMITool):
    def __init__(self, ipmitool_path=DEFAULT_IPMITOOL_PATH):
        super(IPMIToolDrac, self).__init__(ipmitool_path=ipmitool_path)
        self.drac_version = self.get_drac_version()

    def get_drac_version(self):
        out = self.run('sdr elist mcloc', raise_on_error=False)
        if not out.returncode:
            return
        try:
            sensors = [i.split('  ', 1)[0] for i in out]
            version = [i for i in sensors if 'DRAC' in i][0]
        except IndexError:
            return
        return version

    def disable_default_password_prompt(self):
        command = '0x30 0xd0 0x00 0x0a 0x08 0x00 0x00 0x00 0x08 0x00 0x00 0x00 0x00 0x00 0x00 0x00'
        if self.drac_version == 'iDRAC7':
            self.run(self.__raw_command(command))

    def bootstrap_obm_static(self, ip_address, netmask, gateway, username, password):
        self.configure_static_network(ip_address, netmask, gateway)
        self.create_superuser(username, password, user_id=2)
        return self.verify_network_info(ip_address, netmask, gateway, 'static')


class IPMIToolHP(IPMITool):
    def __init__(self,
                 ipmitool_path=DEFAULT_IPMITOOL_PATH,
                 hponcfg_path='/sbin/hponcfg',
                 list_users_path='hpxml/list_users.xml',
                 users_path='hpxml/users.xml',
                 add_user_path='hpxml/add_user.xml',
                 del_user_path='hpxml/del_user.xml'):
        super(IPMIToolHP, self).__init__(ipmitool_path=ipmitool_path)

        script_path = os.path.dirname(__file__)

        def _join(p):
            return os.path.abspath(p) and p or os.path.join(script_path, p)

        self.hponcfg_path = _join(hponcfg_path)
        self.list_users_path = _join(list_users_path)
        self.users_path = _join(users_path)
        self.add_user_path = _join(add_user_path)
        self.del_user_path = _join(del_user_path)

    def configure_static_network(self, ip_address, netmask, gateway, channel=2):
        """
        Default channel is different
        :param ip_address:
        :param netmask:
        :param gateway:
        :param channel:
        :return:
        """
        return super(IPMIToolHP, self).configure_static_network(
            ip_address, netmask, gateway, channel)

    def run_hpon(self, args, bufsize=1048567, raise_on_error=True, ignore_error=False):
        cmd = self.hponcfg_path + ' ' + args
        return cli.run(cmd, bufsize=bufsize, raise_exception=raise_on_error,
                       ignore_error=ignore_error)

    def create_superuser(self, username, password, channels=(1, 2), user_id=2):
        raise NotImplemented

    def create_superuser_hp(self, username, password):
        self.run_hpon('-f %s -l %s' % (self.list_users_path, self.users_path))
        # noinspection PyUnresolvedReferences
        user_tree = etree.parse(self.users_path)
        user_elements = user_tree.xpath('.//USER_LOGIN')
        if user_elements:
            for user_element in user_elements:
                user = list(user_element.values())[0]
                self.run_hpon('-f %s -s user=%s' % (self.del_user_path, user))
        self.run_hpon('-f %s -s user=%s,password=%s' % (self.add_user_path, username, password))

    def bootstrap_obm_static(self, ip_address, netmask, gateway, username, password):
        self.configure_static_network(ip_address, netmask, gateway)
        self.create_superuser_hp(username, password)
        return self.verify_network_info(ip_address, netmask, gateway, 'static')

    def parse_network_info(self, channel=2):
        return super(IPMIToolHP, self).parse_network_info(channel=channel)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
ping / pong payload

{timestamp: 1234567890.12345, message: 'ping'}
{timestamp: 1234567890.12345, load: (m, m5, m15}, message: 'pong'}
"""
import logging
import msgpack
import os
import multiprocessing
import time
import zmq


LOG = logging.getLogger(__name__)


class PongService(object):
    def __init__(self, bind_address):
        self.bind_address = bind_address
        self.ctx = zmq.Context()
        self.socket = self.ctx.socket(zmq.REP)
        self.running = False

    def bind(self):
        LOG.info('Binding pong service: %s' % self.bind_address)
        self.socket.bind(self.bind_address)

    @staticmethod
    def validate(message):
        keys = ['timestamp', 'message']
        for k in keys:
            if k not in message:
                return False
        if message['message'] != 'ping':
            return False

        return True

    def receive(self):
        r = self.socket.recv()

        try:
            message = msgpack.unpackb(r, encoding='utf-8')
        except TypeError as type_error:
            LOG.error('Recieved unpacked, non-string type: %s : %s' % (type(r), type_error))
            return
        except msgpack.UnpackException as unpack_exception:
            LOG.error('Received invalid request: %s' % str(unpack_exception))
            return

        if not self.validate(message):
            LOG.error('Received invalid request: %s' % message)
            return

        LOG.debug('Ping received: %s' % message)

        return message

    def pong(self):
        try:
            loadavg = os.getloadavg()
        except OSError as os_error:
            # Issue: https://github.com/jr0d/mercury/issues/4
            LOG.error('Error getting load average: %s' % str(os_error))
            loadavg = (0.0, 0.0, 0.0)

        _packet = {
            'timestamp': time.time(),
            'load': loadavg,
            'message': 'pong'
        }
        LOG.debug('PONG: %s' % _packet)
        self.socket.send(msgpack.packb(_packet))

    def run(self):
        while True:
            r = self.receive()
            if not r:
                continue

            self.pong()


def _spawn(bind_address):
    ping_service = PongService(bind_address)
    ping_service.bind()
    try:
        ping_service.run()
    except KeyboardInterrupt:
        pass


def spawn_pong_process(bind_address):
    p = multiprocessing.Process(target=_spawn, args=[bind_address])
    p.start()
    LOG.debug('pong service started: PID=%s' % p.pid)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    ps = PongService('tcp://0.0.0.0:9004')
    ps.bind()
    ps.run()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import threading
import time

from mercury.common.clients.rpc.backend import BackEndClient
from mercury.common.exceptions import fancy_traceback_short, parse_exception

log = logging.getLogger(__name__)


class TaskRunner(object):
    """
    """
    def __init__(self, job_id, task_id, entry, backend_url,
                 entry_args=None, entry_kwargs=None, lock=None,
                 timeout=0, task_id_kwargs=False):
        """

        :param job_id:
        :param task_id:
        :param entry:
        :param backend_url:
        :param entry_args:
        :param entry_kwargs:
        :param lock:
        :param timeout:
        :param task_id_kwargs:
        """
        self.job_id = job_id
        self.task_id = task_id
        self.entry = entry
        self.args = entry_args or ()
        self.kwargs = entry_kwargs or {}
        if task_id_kwargs:
            self.kwargs['task_id'] = self.task_id
        self.lock = lock
        self.timeout = timeout

        self.time_started = None
        self.time_completed = None

        self.backend = BackEndClient(backend_url)

    def __management_thread(self):
        """
        Thread to support shared locking.
        :return:
        """
        self.time_started = time.time()

        # Set task to STARTED status
        self.backend.update_task(
            {
                'task_id': self.task_id,
                'status': 'STARTED',
                'time_started': self.time_started
            }
        )

        traceback = None
        # noinspection PyBroadException
        try:
            return_data = self.entry(*self.args, **self.kwargs)
            # TODO: Create response contract for procedures
            if isinstance(return_data, dict) and return_data.get('error'):
                status = 'ERROR'
            else:
                status = 'SUCCESS'
        except Exception:
            exc_dict = parse_exception()
            log.error(fancy_traceback_short(exc_dict,
                                            'Critical error while running task:'
                                            ' %s [%s], elapsed' % (
                                                self.entry.__name__,
                                                self.task_id)),
                      extra={'task_id': self.task_id, 'job_id': self.job_id})
            traceback = parse_exception()
            status = 'ERROR'
            return_data = None
        finally:
            if self.lock:
                log.debug('Releasing lock for %s' % self.lock.task_id,
                          extra={'task_id': self.task_id, 'job_id': self.job_id})
                self.lock.release()

        self.time_completed = time.time()
        log.info('Task completed: %s [%s], elapsed %s' % (self.entry.__name__,
                                                          self.task_id,
                                                          self.time_completed - self.time_started),
                 extra={'task_id': self.task_id, 'job_id': self.job_id})
        log.debug('Publishing response to: %s' % self.backend.zmq_url,
                  extra={'task_id': self.task_id, 'job_id': self.job_id})

        response = self.backend.complete_task({
            'status': status,
            'message': return_data,
            'traceback': traceback,
            'job_id': self.job_id,
            'task_id': self.task_id,
            'time_started': self.time_started,
            'time_completed': self.time_completed,
            'action': 'Completed'
        })
        if response.get('error'):
            log.error('Error dispatching message. [timeout]',
                      extra={'task_id': self.task_id, 'job_id': self.job_id})
        else:
            log.info('Dispatch successful : %s' % response,
                     extra={'task_id': self.task_id, 'job_id': self.job_id})

    def run(self):
        log.info('Starting task: %s [%s]' % (self.entry.__name__, self.task_id),
                 extra={'task_id': self.task_id, 'job_id': self.job_id})
        t = threading.Thread(target=self.__management_thread,
                             name='_{}_{}'.format(self.job_id, self.task_id))
        t.start()



import os
from mercury.common.helpers.cli import run

from mercury_agent.hardware.oem.dell.om_xml_deserializer import (
    XLoader,
    XMLController,
    XMLAbout,
    XMLPDisk,
    XMLVDisk,
    XMLChassisStatus,
)


class OMReportInitException(Exception):
    pass


class OMReportRunException(Exception):
    def __init__(self, message, stdout, stderr):
        self.message = message
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        return repr(self.message)


class OMReport(object):
    def __init__(self, path='/opt/dell/srvadmin/bin/omreport'):
        self.omreport_path = path
        if not os.getuid() == 0:
            raise OMReportInitException('Script must be run as root.')

        if not self._check_service:
            raise OMReportInitException('dataeng services are not running.')

        if not os.path.exists(self.omreport_path):
            raise OMReportInitException('{0} does not exist.'.format(self.omreport_path))

    @property
    def _check_service(self):
        """
        Tested only on redhat/cent
        """
        res = run('service dataeng status')
        if res.returncode:
            return False
        return True

    def run_omreport(self, command, fmt='xml'):
        full_command = '{0} {1} -fmt {2}'.format(self.omreport_path, command, fmt)
        res = run(full_command)
        if res.returncode:
            raise OMReportRunException('Had trouble running {0}'.format(full_command), res,
                                       res.stderr)

        return res.encode()

    @property
    def about(self):
        xml_data = self.run_omreport('about')
        return XMLAbout(XLoader(xml_data).root)

    @property
    def chassis(self):
        xml_data = self.run_omreport('chassis')
        return XMLChassisStatus(XLoader(xml_data).root)

    @property
    def controller(self):
        xml_data = self.run_omreport('controller')
        return XMLController(XLoader(xml_data).root)

    @property
    def vdisk(self):
        xml_data = self.run_omreport('storage vdisk controller=0')
        return XMLVDisk(XLoader(xml_data).root)

    @property
    def pdisk(self):
        xml_data = self.run_omreport('storage pdisk controller=0')
        return XMLPDisk(XLoader(xml_data).root)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Provides functionality for parsing output from `lspci`."""
import shlex
import six
import subprocess

# Class codes used by lspci.
ETHERNET_CONTROLLER = '0200'
NETWORK_CONTROLLER = '0280'
FIBRE_CHANNEL = '0c04'
RAID_CONTROLLER = '0104'


# TODO: Move to mercury.common.exceptions
class LSPCIError(Exception):
    """Raised when something goes wrong related to the `lspci` command."""
    pass


def lspci_run(arguments='-mm'):
    """Runs lspci and returns the output.

    :param arguments: Arguments you want to pass to lspci default = '-mm'
    :return: stdout from lspci command
    :except: LSPCIException on non-zero return code
    """
    cmd = shlex.split('lspci ' + arguments)

    sub_proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    out, err = sub_proc.communicate()

    if sub_proc.returncode:
        raise LSPCIError('[%d] %s' % (sub_proc.returncode, err))

    if isinstance(out, six.binary_type):
        # noinspection PyUnresolvedReferences
        out = out.decode('utf-8')

    return out


def _get_lspci_id(line):
    """Read a hex ID of the form [nnnn] from an `lspci` line.

    Gets an id from a line that looks like this:
        Intel Corporation [8086]
    where 8086 is the id
    It should also work if something like this happens:
    2nd Generation Core Processor Family DRAM [not_and_id] Controller [0104]

    :param line: A string representing a line of lspci output.
    :return: A string representing the ID.
    """
    hush = line.split('[')
    return hush[-1].strip(']')


def _get_lspci_name(line):
    """Reads and returns a 'name' from a line of `lspci` output."""
    hush = line.split('[')
    return '['.join(hush[0:-1]).strip()


def parse_nnvmmk():
    """
    Runs lspci -nnvmmk and parses the output into a list of dictionaries.
    :return: a list of dicts with the following keys
    slot
    vendor_name
    device_name
    svendor_name
    sdevice_name
    vendor_id
    device_id
    svendor_id
    sdevice_id
    revision
    progif
    driver
    :except:
    """
    out = lspci_run('-nnvmmk')
    pcibus = list()

    blocks = out.split('\n\n')

    for block in blocks:
        device = dict()
        for element in block.splitlines():
            split_element = element.split(':')
            key = split_element[0]
            data = ':'.join(split_element[1:]).strip()
            if key in ('Slot', 'ProgIf', 'Driver'):
                device[key.lower()] = data
                continue
            if key in ('Class', 'Vendor', 'Device', 'SVendor', 'SDevice'):
                key_prefix = key.lower()
                device[key_prefix + '_name'] = _get_lspci_name(data)
                device[key_prefix + '_id'] = _get_lspci_id(data)
                continue
            if key == 'Rev':
                device['revision'] = data
                continue
        if not device:
            continue
        pcibus.append(device)

    return pcibus


class PCIDevice(dict):
    """Represents information about a PCI Device as returned by `lspci`."""
    def __init__(self,
                 slot=None,
                 class_id=None,
                 vendor_id=None,
                 device_id=None,
                 class_name='',
                 vendor_name='',
                 device_name='',
                 svendor_name=None,
                 svendor_id=None,
                 sdevice_name=None,
                 sdevice_id=None,
                 revision=None,
                 progif=None,
                 driver=None):
        """Create a PCIDevice. Checks for a few required fields."""
        if None in [slot, class_id, vendor_id, device_id]:
            raise LSPCIError(
                'slot, class_id, vendor_id, and device_id are required.')
        super(PCIDevice, self).__init__()
        self.slot = slot
        self.class_id = class_id
        self.vendor_id = vendor_id
        self.device_id = device_id
        self.class_name = class_name
        self.vendor_name = vendor_name
        self.device_name = device_name
        self.svendor_name = svendor_name
        self.svendor_id = svendor_id
        self.sdevice_name = sdevice_name
        self.sdevice_id = sdevice_id
        self.revision = revision
        self.progif = progif
        self.driver = driver

    def __getattr__(self, key):
        try:
            return self[key]
        except (KeyError, AttributeError):
            return None

    def __setattr__(self, key, value):
        self[key] = value


class PCIBus(list):
    def __init__(self, sudo=False):
        super(PCIBus, self).__init__()
        for it in parse_nnvmmk():
            self.append(PCIDevice(**it))

    def get_devices_by_class(self, class_id):
        sub_li = list()
        for device in self:
            if device.get('class_id') == class_id:
                sub_li.append(device)
        return sub_li

    def has_device_class(self, class_id):
        for device in self:
            if device.get('class_id') == class_id:
                return True
        return False

    def get_devices_by_vendor(self, vendor_id):
        sub_li = list()
        for device in self:
            if device.get('vendor_id') == vendor_id:
                sub_li.append(device)
        return sub_li

    def get_fibre_channel_devices(self):
        return self.get_devices_by_class(FIBRE_CHANNEL)

    def has_fibre_channel(self):
        return self.has_device_class(FIBRE_CHANNEL)

    def get_ethernet_devices(self):
        return self.get_devices_by_class(ETHERNET_CONTROLLER)

    def has_ethernet(self):
        return self.has_device_class(ETHERNET_CONTROLLER)

    def get_network_devices(self):
        return self.get_devices_by_class(NETWORK_CONTROLLER)

    def has_network(self):
        return self.has_device_class(NETWORK_CONTROLLER)

    def get_raid_bus_controller_devices(self):
        return self.get_devices_by_class(RAID_CONTROLLER)

    def has_raid_bus_controller(self):
        return self.has_device_class(RAID_CONTROLLER)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from press.helpers import parted

from mercury_agent.inspector.inspectors import inspector
from mercury_agent.inspector.hwlib.udev import UDevHelper

log = logging.getLogger(__name__)


def get_disk_type(dev):
    dev = dev.lstrip('/dev/')
    sysfs_rotational_path = f'/sys/block/{dev}/queue/rotational'
    try:
        return 'disk' if int(open(sysfs_rotational_path).read()) else 'ssd'
    except (IOError, OSError):
        log.warning('Could not parse sysfs for %s', dev)
        return 'unknown'


@inspector.expose('os_storage')
def os_storage_inspector():
    os_storage = []
    storage_devices = UDevHelper().discover_valid_storage_devices(
        fc_enabled=True, loop_enabled=False)
    for storage_device in storage_devices:
        device_info = {}
        udev_device = dict(list(storage_device.items()))
        # Requires root privilages
        try:
            device_info.update(**parted.PartedInterface(
                udev_device['DEVNAME']).device_info)
        except parted.PartedException:
            log.warning(
                'Could not parse disk label for %s. Got root?',
                udev_device['DEVNAME'])
        device_info['udev'] = udev_device
        if udev_device.get('ID_BUS') != 'fc':
            device_info['media_type'] = get_disk_type(udev_device['DEVNAME'])
        else:
            # Fibre channel!
            device_info['media_type'] = 'external'
        os_storage.append(device_info)
    return os_storage


if __name__ == '__main__':
    import json
    print(json.dumps(os_storage_inspector(), indent=2))



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import netifaces


def get_default_interface():
    """Get default AF_INET interfaces.

    :return: kernel device name or '' if no default interface found.
    """
    gws = netifaces.gateways()
    if 'default' not in gws:
        return ''

    route = gws['default'].get(netifaces.AF_INET)
    if route:
        return route[1]


def get_link_addresses(exclude_loopback=True):
    """Format a list containing AF_LINK info for each interface.

    :param exclude_loopback: Bool
    :return: list of dicts : {
        'interface': interface_name,
        'mac_address': ethernet address
    }
    """
    interfaces = list_interfaces(exclude_loopback)
    link_info = []
    for interface in interfaces:
        addresses = netifaces.ifaddresses(interface)
        link_address = addresses.get(netifaces.AF_LINK)
        if link_address:
            link_info.append({
                'interface': interface,
                'mac_address': link_address[0]['addr']
            })

    return link_info


def get_ipv4_network_info(interface):
    """Helper to get only AF_INET addresses.

    Be aware this function will return a list of addresses associated
    with the interface. The order of the list is unknown and matching
    networks to routes may be required in instances where there are multiple
    addresses (vlans, aliases, etc). As such, our code should check the
    length of the list even when we assume a length of 1.

    :param interface: device name, eth0, enp3s0, em1, etc.
    :return: A list of ipv4 info dicts {addr, broadcast || peer, netmask}
    """
    if interface not in list_interfaces():
        return []
    addresses = netifaces.ifaddresses(interface)
    return addresses.get(netifaces.AF_INET, [])


def get_ipv6_network_info(interface):
    """The same as get_ipv4_network_info but for AF_INET6

    :param interface: device name, eth0, enp3s0, em1, etc.
    :return: A list of ipv6 info dicts {addr, broadcast || peer, netmask}
    """
    if interface not in list_interfaces():
        return []
    addresses = netifaces.ifaddresses(interface)
    return addresses.get(netifaces.AF_INET6, [])


def list_interfaces(exclude_loopback=True):
    """Helper function to get network interfaces.

    :param exclude_loopback: If True, then the loopback interface is not
        included in the returned list. If False, the loopback interface will
        be included.
    :return: A list of interfaces.
    """
    interfaces = netifaces.interfaces()
    if exclude_loopback:
        interfaces = [i for i in interfaces if i != 'lo']

    return interfaces



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import threading

from mercury_agent.capabilities import runtime_capabilities
from mercury_agent.task_runner import TaskRunner
from mercury.common.transport import SimpleRouterReqService

LOG = logging.getLogger(__name__)


class SerialLock(object):
    def __init__(self):
        self.task_id = None
        self.lock = threading.Lock()

    def acquire(self, task_id):
        self.task_id = task_id
        return self.lock.acquire(False)  # don't block on acquire

    def release(self):
        self.task_id = None
        self.lock.release()


class AgentService(SimpleRouterReqService):
    serial_lock = SerialLock()

    def __init__(self, bind_address, rpc_backend_url):
        """The Agent RPC service
        :param bind_address: The the zeromq bind address for the service
        :param rpc_backend_url: RPC backend URL
        """
        self.rpc_backend_url = rpc_backend_url
        super(AgentService, self).__init__(bind_address)

    @staticmethod
    def error(code, data=''):
        return {'message': {'status': code, 'data': data}}

    @staticmethod
    def sync_response(data):
        return {'message': {'status': 0, 'data': data}}

    @staticmethod
    def lookup_method(method):
        return runtime_capabilities.get(method)

    def validate_method_args(self, message, capability):
        num_args = capability.get('num_args')
        if num_args:
            if not len(message.get('args', 0)) == num_args:
                return self.error(4,
                                  'arguments for %s are not satisfied' %
                                  capability['name'])

        kwarg_names = capability.get('kwarg_names')
        if kwarg_names:
            message_kwargs = message.get('kwargs', {})
            if not isinstance(message_kwargs, dict):
                # could happen if message.kwargs is explicitly set to something other than a dict
                return self.error(4, 'kwargs is malformed in message')
            missing = []
            for k in kwarg_names:
                if k not in message_kwargs:
                    missing.append(k)
            if missing:
                return self.error(4, 'missing kwargs in message: %s' % ', '.join(missing))

    def process(self, message):
        LOG.debug('Processing request: %s' % message)

        category = message.get('category')
        if not category:
            return self.error(1, 'missing category, nothing to do')

        if category != 'rpc':
            return self.error(3, 'invalid category')

        method = message.get('method')
        if not method:
            return self.error(3, 'method not defined')

        capability = self.lookup_method(method)
        if not capability:
            return self.error(3, '%s method is not supported' % method)

        LOG.debug('Runtime selected: %s' % capability['name'])

        error = self.validate_method_args(message, capability)
        if error:
            return error

        args = message.get('args', ())
        kwargs = message.get('kwargs', {})

        task_id = message.get('task_id')
        job_id = message.get('job_id')

        if None in [task_id, job_id]:
            return self.error(3, 'message is incomplete, missing task_id/job_id')

        # Task runner inject
        # ret = capability['entry'](*args, **kwargs)

        if capability.get('serial'):
            LOG.info('capability %s is serial, attempting to acquire lock' %
                     capability['name'])
            if not self.serial_lock.acquire(task_id):
                LOG.info('Could not acquire lock, locked by %s' % task_id)
                return self.error(5, 'Could not acquire lock, task is already running')
            LOG.info('Lock acquired for %s' % task_id)
            tmp_lock = self.serial_lock
        else:
            tmp_lock = None  # We don't care about the lock, unless we are running
            # a serial task

        task_runner = TaskRunner(job_id,
                                 task_id,
                                 capability['entry'],
                                 self.rpc_backend_url,
                                 entry_args=args,
                                 entry_kwargs=kwargs,
                                 lock=tmp_lock,
                                 timeout=capability['timeout'],
                                 task_id_kwargs=capability['task_id_kwargs'])
        task_runner.run()
        return self.sync_response(data=dict(time_started=task_runner.time_started,
                                            message='Hakuna matata'))



import logging
import threading

from mercury.common.exceptions import MercuryGeneralException
from mercury.common.helpers import cli

log = logging.getLogger(__name__)
LLDPLITE_DEFAULT_PATH = 'lldplite'


class LLDPInspector(object):
    __key__ = 'lldplite'

    def __init__(self, device_info, backend_client, lldplite_path=LLDPLITE_DEFAULT_PATH):
        self.device_info = device_info
        self.backend_client = backend_client
        self.lldplite_path = lldplite_path
        self.lldplite = cli.find_in_path(lldplite_path)
        if not self.lldplite:
            raise MercuryGeneralException('Could not find lldplite binary')

        self.pids = {}

    def _run(self, interface):
        switch_info = get_lldp_info(interface, self.lldplite)
        self.process(interface, switch_info)

    def inspect(self):
        for i in self.device_info.get('interfaces', []):
            if i.get('carrier'):
                log.info('Interface {} is active, sniffing for LLDP packets'.format(i['devname']))
                p = threading.Thread(target=self._run, args=(i['devname'],))
                self.pids[i['devname']] = {'process': p, 'result': None}
                p.start()

    def get_interface_index(self, interface):
        interfaces = self.device_info.get('interfaces')
        if interfaces:
            for _i in range(len(interfaces)):
                if interfaces[_i]['devname'] == interface:
                    return _i
        log.error('{} index is somehow missing from device info'.format(interface))
        return -1

    def process(self, interface, switch_info):
        index = self.get_interface_index(interface)

        if not switch_info or index == -1:
            self.pids[interface]['result'] = {'missing': True}
            return

        switch_info['interface_index'] = self.get_interface_index(interface)
        self.backend_client.update(self.device_info['mercury_id'], {'lldp': switch_info})
        self.pids[interface]['result'] = switch_info

    def cleanup(self):
        pass


def get_lldp_info(interface, lldplight_path):
    command = '{} {}'.format(lldplight_path, interface)
    result = cli.run(command, raise_exception=False, quiet=True, ignore_error=True)

    if result.returncode == 5:
        log.debug('Timed out waiting for LLDP broadcast on {}'.format(interface))
        return None

    elif result.returncode:
        log.error("Problem running lldplite: {}".format(result.stderr))
        return None

    switch_name = result.split()[0]
    port_number = int(result.split()[1].split('/')[-1])

    log.info('LLDP info: {} {}'.format(switch_name, port_number))

    return {'switch_name': switch_name, 'port_number': port_number}



RAID_CONTROLLER_CLASS_ID = "0104"
SMART_ARRAY_DEVICE_ID_9 = "3239"  # Smart Array Gen9 Controllers


def get_vendor(dmi_info):
    return dmi_info.get('sys_vendor')


def is_platform(dmi_info, platform):
    return get_vendor(dmi_info) == platform


def is_dell(dmi_info):
    return is_platform(dmi_info, 'Dell Inc.')


def is_hp(dmi_info):
    return is_platform(dmi_info, 'HP')


def is_quanta(dmi_info):
    return is_platform(dmi_info, 'Quanta')


def get_product_name(dmi_info):
    return dmi_info.get('product_name')


# PCI
def has_smart_array_gen9(pci_data):
    for device in pci_data:
        if device['device_id'] == SMART_ARRAY_DEVICE_ID_9:
            return True
    return False


def get_raid_controllers(pci_data):
    return [d for d in pci_data if d['class_id'] == RAID_CONTROLLER_CLASS_ID]
